'''
Sets of function to compute hydrodynamics with RiverDale

B.G. - 29/04/2024
'''

import taichi as ti
import numpy as np
from enum import Enum
import scabbard.utils as scaut 
from scabbard.riverdale.rd_grid import GRID
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.riverdale.rd_helper_surfw as hsw
import scabbard.riverdale.rd_utils as rut



class FlowMode(Enum):
	'''
	Enumeration of the different boundary condition types possible
	'''	
	static_incremental = 0
	static_drape = 1
	static_link = 2


@scaut.singleton
class HydroParams:
	'''
		Internal singleton class holding all the compile time constant parameters for the hydro simulations 
		Not for users
	'''
	def __init__(self):		
		# Constant dt for hydrodynamics
		self.dt_hydro = 1e-3
		# Constant manning coefficient
		self.manning = 0.033
		self.exponent_flow = np.float32(5./3.)
		self.flowmode = FlowMode.static_incremental
		self.hydro_slope_bc_mode = 0
		self.hydro_slope_bc_val = 0

		# Constant for minimal step in the draping function
		self.mini_drape_step = np.float32(0.0005)

		self.use_heffmax = False

		self.use_original_dir_for_LM = True
		self.LM_pathforcer = 1
		self.clamp_div_hw = True
		self.clamp_div_hw_val = 1e-3

		self.checkboard_checker = False

		self.CA_redis = 0.1

PARAMHYDRO = HydroParams()

@ti.kernel
def initiate_step(QwB: ti.template()):
	'''
	Runs the initial operations when running a step
	Mostly a placeholder so far but keeps things consistents
	Arguments:
		- QwB: a 2D field of discharge B (in_temp)
	Returns:
		- Nothing, edits in place
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	# Reinitialise QwB to 0.
	for i,j in QwB:
		QwB[i,j] = 0.


@ti.kernel
def constant_rain(QwA: ti.template(), QwB: ti.template(), P: ti.f32, BCs:ti.template()):
	'''
	Adds a constant precipitation rates to every cells
	Arguments:
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in_temp)
		- P: a constant floating point of precipitation rates in m/s
	Returns:
		- Nothing, edits in place
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	for i,j in QwA:

		if(gridfuncs.is_active(i,j, BCs) == False or gridfuncs.can_give(i,j,BCs) == False):
			continue
		
		# QwA[i,j] += P * GRID.dx * GRID.dy
		QwB[i,j] = P * GRID.dx * GRID.dy

@ti.kernel
def variable_rain(QwA: ti.template(), QwB: ti.template(), P: ti.template(), BCs:ti.template()):
	'''
	Adds a spatially variable precipitation rates to every cells
	Arguments:
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in_temp)
		- P: a 2D field of precipitation rates in m/s
	Returns:
		- Nothing, edits in place
	Authors:
		- B.G. (last modification 30/04/2024)
	'''
	for i,j in QwA:

		if(gridfuncs.is_active(i,j, BCs) == False or gridfuncs.can_give(i,j,BCs) == False):
			continue

		# QwA[i,j] += P[i,j] * GRID.dx * GRID.dy
		QwB[i,j] = P[i,j] * GRID.dx * GRID.dy

@ti.kernel
def input_discharge_points(input_rows: ti.template(), input_cols:ti.template(), input_values:ti.template(), QwA: ti.template(), QwB: ti.template(), BCs:ti.template()):
	'''
	Adds discharge in m^3/s into specific input points
	Arguments:
		- input_rows: a 1D field of input point row coordinates (integrer 32 bits)
		- input_cols: a 1D field of input point column coordinates (integrer 32 bits)
		- input_values: a 1D field of input discharges in m^3/s (floating point)
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in_temp)
	Returns:
		- Nothing, edits in place
	Authors:
		- B.G. (last modification 30/04/2024)
	'''

	for i in input_rows:
		# QwA[input_rows[i],input_cols[i]] += input_values[i]
		QwB[input_rows[i],input_cols[i]] += input_values[i]



@ti.kernel
def _compute_Qw(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template(), flowdir:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
		- fdir: a 2D field of flow direction to follow in the event of a local minima
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue

		if(QwA[i,j] == 0 and hw[i,j] == 0):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		# allmore = True
		# allless = True

		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):
				
			# Traversing Neighbours
			for k in range(4):

				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				if(gridfuncs.can_receive(ir,jr, BCs) == False):
					continue

				# Local hydraulic slope
				tS = hsw.Sw(Z,hw,i,j,ir,jr)

				# If < 0, neighbour is a donor and I am not interested
				if(tS <= 0):
					continue

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSy):
						SSy = tS
				else:
					if(tS > SSx):
						SSx = tS

				# Registering local slope
				Sws[k] = tS
				# Summing it to global
				sumSw += tS

				# if(QwA[ir,jr] < QwA[i,j]):
				# 	allmore = False
				# elif(QwA[ir,jr] > QwA[i,j]):
				# 	allless = False

				# Done with processing this particular neighbour

			# Local minima management (cheap but works)
			## If I have no downward slope, I increase the elevation by a bit
			if(sumSw == 0.): # or ((allmore or allless) and PARAMHYDRO.checkboard_checker)):
				# I am in a local minima
				# this option ensures the drainage of LM through the original rail
				# Flow dir == 5 is no flow
				if(flowdir[i,j] != 5):
					# That section ensures that a stochastic number of receivers are traversed to flush the local minima
					# And avoid ping-pong or localisation based biases
					ii,jj = i,j
					ir,jr = i,j
					first = 0
					# Receivers are poped out at least once, and then has a probability of 0.5 to continue
					while(flowdir[ir,jr] != 5):
						if((first>PARAMHYDRO.LM_pathforcer) and ti.random() < 0.5):
							break
						ii,jj = ir,jr
						first += 1
						ir,jr = gridfuncs.neighbours(ii, jj, flowdir[ii,jj], BCs)
						# ti.atomic_add(QwB[ir,jr], QwA[i,j])
					ti.atomic_add(QwB[ir,jr], QwA[i,j])
				else:
					ti.atomic_add(QwB[i,j], QwA[i,j])

				QwC[i,j] = 0.
				continue

			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			thw = ti.math.max(0., hw[i,j])

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(thw, PARAMHYDRO.exponent_flow) * sumSw/ti.math.sqrt(gradSw)

				
			# Transferring flow to neighbours
			for k in range(4):

				# local neighbours
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)
				
				# checking if neighbours
				if(ir == -1):
					continue
				
				# Transferring prop to the hydraulic slope
				ti.atomic_add(QwB[ir,jr], Sws[k]/sumSw * QwA[i,j])

		# Boundary case
		else:
			tSw = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val
			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * ti.math.sqrt(tSw)

@ti.kernel
def _compute_Qw_dynamic(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template()):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
		- fdir: a 2D field of flow direction to follow in the event of a local minima
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):
				
			# Traversing Neighbours
			for k in range(4):

				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				if(gridfuncs.can_receive(ir,jr, BCs) == False):
					continue

				# Local hydraulic slope
				tS = hsw.Sw(Z,hw,i,j,ir,jr)

				# If < 0, neighbour is a donor and I am not interested
				if(tS <= 0):
					continue

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSy):
						SSy = tS
				else:
					if(tS > SSx):
						SSx = tS

				# Registering local slope
				Sws[k] = tS
				# Summing it to global
				sumSw += tS

				# Done with processing this particular neighbour

			# No receivers, no output discharge
			if(sumSw == 0.):
				QwC[i,j] = 0.
				QwB[i,j] = QwA[i,j]
				continue

			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			thw = ti.math.max(0., hw[i,j])

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(thw, PARAMHYDRO.exponent_flow) * sumSw/ti.math.sqrt(gradSw)

				
			# Transferring flow to neighbours
			for k in range(4):

				# local neighbours
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)
				
				# checking if neighbours
				if(ir == -1):
					continue
				
				# Transferring prop to the hydraulic slope
				ti.atomic_add(QwB[ir,jr], Sws[k]/sumSw * QwC[i,j])

		# Boundary case
		else:
			tSw = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val
			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * ti.math.sqrt(tSw)





@ti.kernel
def _compute_hw(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template() ):
	'''
	Compute flow depth by div.Q and update discharge to t+1.
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, update flow depth in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	
	# print(PARAMHYDRO.dt_hydro,"OIOIOIOI")
	# Traversing nodes
	for i,j in Z:
		
		# Updating local discharge to new time step
		QwA[i,j] = QwB[i,j]
		
		# ONGOING TEST DO NOT DELETE
		# # Only where nodes are active (i.e. flow cannot leave and can traverse)
		# if(gridfuncs.can_out(i,j,BCs)):
		# 	continue


		# Updating flow depth (cannot be < 0)
		dhw = (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy)

		if(PARAMHYDRO.clamp_div_hw):
			if(dhw>0):
				dhw = ti.math.min(PARAMHYDRO.clamp_div_hw_val, dhw)
			else:
				dhw = ti.math.max(-PARAMHYDRO.clamp_div_hw_val, dhw)

		hw[i,j] = ti.max(0.,hw[i,j] + dhw)

@ti.kernel
def _compute_hw_dynamic(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template() ):
	'''
	Compute flow depth by div.Q and update discharge to t+1.
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, update flow depth in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	
	# print(PARAMHYDRO.dt_hydro,"OIOIOIOI")
	# Traversing nodes
	for i,j in Z:
		
		
		
		# ONGOING TEST DO NOT DELETE
		# # Only where nodes are active (i.e. flow cannot leave and can traverse)
		# if(gridfuncs.can_out(i,j,BCs)):
		# 	continue


		# Updating flow depth (cannot be < 0)
		dhw = (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy)

		if(PARAMHYDRO.clamp_div_hw):
			if(dhw>0):
				dhw = ti.math.min(PARAMHYDRO.clamp_div_hw_val, dhw)
			else:
				dhw = ti.math.max(-PARAMHYDRO.clamp_div_hw_val, dhw)

		hw[i,j] = ti.max(0.,hw[i,j] + dhw)

		# Updating local discharge to new time step
		QwA[i,j] = QwB[i,j]


@ti.kernel
def _raise_analytical_hw(Z:ti.template(), hw:ti.template(), QwA:ti.template(), temp:ti.template(), BCs:ti.template()):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''


	# Traversing each nodes
	for i,j in Z:

		temp[i,j] = hw[i,j]

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		# Safety check: gets incremented at each while iteration and manually breaks the loop if > 10k (avoid getting stuck in an infinite hole)
		lockcheck = 0

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue

		thw = 0.

		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):

			# Traversing Neighbours
			for k in range(4):

				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				if(gridfuncs.can_receive(ir,jr, BCs) == False):
					continue

				# Local hydraulic slope
				tS = hsw.Sw(Z,hw,i,j,ir,jr)

				# If < 0, neighbour is a donor and I am not interested
				if(tS <= 0):
					continue

				thw = ti.math.max(thw, ti.max(hsw.Zw(Z,hw,i,j),hsw.Zw(Z,hw,ir,jr)) - ti.max(Z[i,j],Z[ir,jr]))

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSy):
						SSy = tS
				else:
					if(tS > SSx):
						SSx = tS


			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy) if SSx > 0 or SSy > 0 else 1e-5

			temp[i,j] = (QwA[i,j] * PARAMHYDRO.manning/(GRID.dx * (gradSw)**(0.5) ))**(3./5.)

	for i,j in Z:
		hw[i,j] = temp[i,j]


@ti.kernel
def _CA_smooth(Z:ti.template(), hw:ti.template(), hw_new:ti.template(), BCs:ti.template()):

	for i,j in Z:
		hw_new[i,j] = hw[i,j]
	
	for i,j in Z:
	
		if(gridfuncs.can_out(i, j, BCs) or gridfuncs.is_active(i,j,BCs) == False):
			continue

		totS = 0.
		dh = PARAMHYDRO.CA_redis * hw[i,j]

		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)
			if(ir == -1):
				continue

			tS = (Z[i,j] - Z[ir,jr] + hw[i,j] - hw[ir,jr])/GRID.dx
			if tS > 0.:
				totS += tS

		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)
			if(ir == -1):
				continue
			tS = (Z[i,j] - Z[ir,jr] + hw[i,j] - hw[ir,jr])/GRID.dx
			if tS > 0.:
				hw_new[ir,jr] += tS * dh / totS

		hw_new[i,j] -= dh

	for i,j in Z:
		hw[i,j] = hw_new[i,j]


@ti.kernel
def _compute_Qw_surfrec(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template(), surfrec:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		# Safety check: gets incremented at each while iteration and manually breaks the loop if > 10k (avoid getting stuck in an infinite hole)
		lockcheck = 0

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue

		thw = 0.

		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):
			# While I do not have external slope
		
			# Traversing Neighbours
			for k in range(4):

				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				if(gridfuncs.can_receive(ir,jr, BCs) == False):
					continue

				if(surfrec[i,j] <= surfrec[ir,jr]):
					continue

				# Local hydraulic slope
				tS = hsw.Sw(Z,hw,i,j,ir,jr)

				if(tS <= 0.):
					tS = 1e-3

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSy):
						SSy = tS
				else:
					if(tS > SSx):
						SSx = tS

				# Registering local slope
				Sws[k] = tS
				# Summing it to global
				sumSw += tS

				# Done with processing this particular neighbour

		
			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			
			# Not sure I still need that
			if(gradSw == 0):
				continue

			thw = ti.max(hw[i,j], 0.)

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(thw, PARAMHYDRO.exponent_flow) * sumSw/ti.math.sqrt(gradSw)


			# Transferring flow to neighbours
			for k in range(4):

				# local neighbours
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)
				
				# checking if neighbours
				if(ir == -1):
					continue
				
				# Transferring prop to the hydraulic slope
				ti.atomic_add(QwB[ir,jr], Sws[k]/sumSw * QwA[i,j])

		# Boundary case
		else:
			tSw = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val
			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * ti.math.sqrt(tSw)
			
			

@ti.kernel
def _compute_hw_CFL(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template(), alpha : ti.f32, threshold:ti.f32 ):
	'''
	Status: STILL EXPERIMENTAL
	Variant of _compute_hw that does not consider nodes nearly equillibrated AND
	
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, update flow depth in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	

	# Traversing nodes
	for i,j in Z:
		
		# Updating local discharge to new time step
		QwA[i,j] = QwB[i,j]
		
		# ONGOING TEST DO NOT DELETE
		# # Only where nodes are active (i.e. flow cannot leave and can traverse)
		# if(gridfuncs.can_out(i,j,BCs)):
		# 	continue

		if(QwA[i,j] <= 0 or abs(1 - (QwC[i,j]/QwA[i,j])) < threshold ):
			continue


		# Updating flow depth (cannot be < 0)
		tdt = PARAMHYDRO.dt_hydro
		if(QwC[i,j] > 0):	
			tdt = ti.math.max(PARAMHYDRO.dt_hydro, alpha *GRID.dx/(QwA[i,j]/(ti.math.max(1e-4, hw[i,j]))) )
			
		hw[i,j] = hw[i,j] + (QwA[i,j] - QwC[i,j]) * tdt/(GRID.dx*GRID.dy) 

@ti.kernel
def _compute_hw_th(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template(), tdt:ti.f32, threshold:ti.f32 ):
	'''
	Compute flow depth by div.Q and update discharge to t+1.
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, update flow depth in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	

	# Traversing nodes
	for i,j in Z:
		
		# Updating local discharge to new time step
		QwA[i,j] = QwB[i,j]
		
		# ONGOING TEST DO NOT DELETE
		# # Only where nodes are active (i.e. flow cannot leave and can traverse)
		# if(gridfuncs.can_out(i,j,BCs)):
		# 	continue

		if(QwA[i,j] <= 0 or abs(1 - (QwC[i,j]/QwA[i,j])) < threshold ):
			continue


		# Updating flow depth (cannot be < 0)
		# tdt = ti.math.max(PARAMHYDRO.dt_hydro, alpha *GRID.dx/(QwA[i,j]/(ti.math.max(1e-4, hw[i,j]))) )		
		hw[i,j] = hw[i,j] + (QwA[i,j] - QwC[i,j]) * tdt/(GRID.dx*GRID.dy) 


@ti.kernel
def _flush_QwA_only(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), BCs:ti.template(), flowdir:ti.template() ):
	'''
	Compute one iteration of QwA propagation

	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''

	for i,j in Z:
		QwB[i,j] = 0.


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue


		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):


			if(flowdir[i,j] != 5):
				ir,jr = gridfuncs.neighbours(i,j,flowdir[i,j], BCs)
				# Transferring prop to the hydraulic slope
				ti.atomic_add(QwB[ir,jr], QwA[i,j])

	for i,j in Z:
		QwA[i,j] = QwB[i,j]



@ti.kernel
def _propagate_QwA_only(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), BCs:ti.template() ):
	'''
	Compute one iteration of QwA propagation

	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''

	for i,j in Z:
		QwB[i,j] = 0.


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue


		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):


			# Traversing Neighbours
			for k in range(4):

				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				if(gridfuncs.can_receive(ir,jr, BCs) == False):
					continue

				# Local hydraulic slope
				tS = hsw.Sw(Z,hw,i,j,ir,jr)

				# If < 0, neighbour is a donor and I am not interested
				if(tS <= 0):
					continue

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSy):
						SSy = tS
				else:
					if(tS > SSx):
						SSx = tS

				# Registering local slope
				Sws[k] = tS
				# Summing it to global
				sumSw += tS

				# Done with processing this particular neighbour

			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			
			# Not sure I still need that
			if(gradSw == 0):
				continue

			# Transferring flow to neighbours
			for k in range(4):

				# local neighbours
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)
				
				# checking if neighbours
				if(ir == -1):
					continue
				
				# Transferring prop to the hydraulic slope
				ti.atomic_add(QwB[ir,jr], Sws[k]/sumSw * QwA[i,j])

	for i,j in Z:
		QwA[i,j] = QwB[i,j]


@ti.kernel
def _compute_QwA_from_Zw(Z:ti.template(), hw:ti.template(), QwA:ti.template(), BCs:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue

		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):
							


			# Traversing Neighbours
			for k in range(4):

				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				if(gridfuncs.can_receive(ir,jr, BCs) == False):
					continue

				# Local hydraulic slope
				tS = hsw.Sw(Z,hw,i,j,ir,jr)

				# If < 0, neighbour is a donor and I am not interested
				if(tS <= 0):
					continue

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSy):
						SSy = tS
				else:
					if(tS > SSx):
						SSx = tS

				# Registering local slope
				Sws[k] = tS
				# Summing it to global
				sumSw += tS

				# Done with processing this particular neighbour

				# Local minima management (cheap but works)
				## If I have no downward slope, I increase the elevation by a bit
				if(sumSw == 0.):
					QwA[i,j] = 0.


			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			
			# Not sure I still need that
			if(gradSw == 0):
				QwA[i,j] = 0.
				continue

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwA[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * sumSw/ti.math.sqrt(gradSw)

		# Boundary case
		else:
			tSw = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val
			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwA[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * ti.math.sqrt(tSw)

@ti.kernel
def _compute_Qw_drape(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template() ):
	'''
	EXPERIMENTAL: testing some dynamic draping
	Variant of Qw that applies a draping algorithm to avoid the creation of local minimas
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		# Safety check: gets incremented at each while iteration and manually breaks the loop if > 10k (avoid getting stuck in an infinite hole)
		lockcheck = 0


		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):
			# First incrementing the safety check
			lockcheck += 1

			# Traversing Neighbours
			for k in range(4):

				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				if(gridfuncs.can_receive(ir,jr, BCs) == False):
					continue

				# Local hydraulic slope
				tS = hsw.Sw(Z,hw,i,j,ir,jr)

				# If < 0, neighbour is a donor and I am not interested
				if(tS <= 0):
					continue

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(k == 0 or k == 3):
					if(tS > SSy):
						SSy = tS
				else:
					if(tS > SSx):
						SSx = tS

				# Registering local slope
				Sws[k] = tS
				# Summing it to global
				sumSw += tS

				# Done with processing this particular neighbour

			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			
			# Not sure I still need that
			if(gradSw == 0):
				continue

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * sumSw/ti.math.sqrt(gradSw)

			# Transferring flow to neighbours
			for k in range(4):

				# local neighbours
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)
				
				# checking if neighbours
				if(ir == -1):
					continue
				
				# Transferring prop to the hydraulic slope
				ti.atomic_add(QwB[ir,jr], Sws[k]/sumSw * QwA[i,j])

		# Boundary case
		else:
			tSw = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val
			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * ti.math.sqrt(tSw)


@ti.kernel
def _compute_hw_drape(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), constrains:ti.template(), BCs:ti.template() ):
	'''
	EXPERIMENTAL
	Compute flow depth by div.Q and update discharge to t+1.
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- constrains: a 3D field of minimum [i,j,0] and maximum [i,j,1] Zw possible for every nodes
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, update flow depth in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	

	#Reinitialising the constrains to their default values
	for i,j in Z:

		constrains[i,j,0] = -1e6
		constrains[i,j,1] = 1e6

	# Constant of draping, manages
	rat = ti.f32(0.45)

	# First determining the draping constrains
	for i,j in Z:

		# Local elevation
		tZw = hsw.Zw_drape(Z,hw,i,j)

		# Ignoring nodes outletting (they still get a constrain[i,j,1] given by donors)
		if gridfuncs.can_out(i,j,BCs):
			continue

		# keeping track of the mini elev
		Zdkmin = tZw
		kmin = 5

		# Traversing Neighbours
		for k in range(4):

			# Getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# not a neighbour
			if(ir == -1):
				continue

			# Note that Zw_drape() returns the absolute hw + Z, even if hw is < 0
			tZwr = hsw.Zw_drape(Z,hw,ir,jr)

			# Registering new steepest neighbour if needed
			if(tZwr < Zdkmin):
				Zdkmin = tZwr
				kmin = k

		# Local minima ? I skip
		if kmin == 5:
			continue

		# Row col of the receiver
		ir,jr = gridfuncs.neighbours(i,j,kmin, BCs)

		# Calculating constrains
		if(gridfuncs.can_receive(ir,jr,BCs)):

			# Checking if I have enough space between my nodes
			if(tZw - Zdkmin >= PARAMHYDRO.mini_drape_step):

				# if I do, I calculate the min/max function of a slightly unbalanced mean between our node and its receiver
				constrains[i,j,0] = ti.math.max(rat * (hsw.Zw_drape(Z,hw,ir,jr)) + (1 - rat) * (tZw), hsw.Zw_drape(Z,hw,ir,jr) + PARAMHYDRO.mini_drape_step)

				# Propagating to tneighbour's max possible elevation
				ti.atomic_min(constrains[ir,jr,1] , ti.math.min( (1-rat) * hsw.Zw_drape(Z,hw,ir,jr) + rat * tZw, tZw - PARAMHYDRO.mini_drape_step) )

			else:
				# If we do not have enough space
				# i "lock" the node for now. It will be eventually unlocked by reorganisation of neighbours
				constrains[i,j,0] = tZw

				# Propagating to tneighbour's max possible elevation
				ti.atomic_min(constrains[ir,jr,1] , hsw.Zw_drape(Z,hw,ir,jr))


	# Draped increment 
	for i,j in Z:
		

		# Updating local discharge to new time step
		QwA[i,j] = QwB[i,j]
				
		# Applying the increment whatsoever YOLO lol
		hw[i,j] = hw[i,j] + (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy) 
		
		# Converting constrains into flow depth clamper
		constrains[i,j,0] -= Z[i,j]
		constrains[i,j,1] -= Z[i,j]


		# Clamp
		if(hw[i,j] < constrains[i,j,0]):
			hw[i,j] = constrains[i,j,0]
		elif(hw[i,j] > constrains[i,j,1]):
			hw[i,j] = constrains[i,j,1]



@ti.kernel
def _compute_hw_drape_th(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), D4dir:ti.template(), constrains:ti.template(), BCs:ti.template(), threshold:ti.f32 ):
	'''
	EXPERIMENTAL
	Compute flow depth by div.Q and update discharge to t+1.
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- constrains: a 3D field of minimum [i,j,0] and maximum [i,j,1] Zw possible for every nodes
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, update flow depth in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	

	for i,j in Z:

		constrains[i,j,0] = -1e6
		constrains[i,j,1] = 1e6
		D4dir[i,j] = 5

	rat = 0.45

	for i,j in Z:

		tZw = hsw.Zw_drape(Z,hw,i,j) # + (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy)
		if gridfuncs.is_active(i,j,BCs) == False:
			continue

		kmin = 5
		Zdkmin = tZw

		# Traversing Neighbours
		for k in range(4):

			# Getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			if(ir == -1):
				continue

			# tSwr = hsw.Zw_drape(Z,hw,ir,jr) + (QwA[ir,jr] - QwC[ir,jr]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy)
			tZwr = hsw.Zw_drape(Z,hw,ir,jr)

			# if(tZwr == tZw):
			# 	print('HAPPENS:',i,j,'vs',ir,jr, tZwr,tZw)

			if(tZwr < Zdkmin):
				Zdkmin = tZwr
				kmin = k

		if kmin==5:
			continue

		ir,jr = gridfuncs.neighbours(i,j,kmin, BCs)


		D4dir[i,j] = kmin

		# rat = 0.45 + ti.random() * 0.04
		# rat = 1.

		constrains[i,j,0] = (rat * (hsw.Zw_drape(Z,hw,ir,jr)) + (1 - rat) * (tZw))
		# constrains[i,j,0] = hsw.Zw_drape(Z,hw,ir,jr)


		ti.atomic_min(constrains[ir,jr,1] , ((1-rat) * (hsw.Zw_drape(Z,hw,ir,jr)) + rat * (tZw)))

		# ti.atomic_min(constrains[ir,jr,1] , tZw)

		# constrains[i,j,0] -= Z[i,j]
		# constrains[i,j,1] -= Z[i,j]

	# for i,j in Z:
	# 	if(constrains[i,j,0] == constrains[i,j,1]):
	# 		ir,jr = gridfuncs.neighbours(i,j,D4dir[i,j], BCs)
	# 		print('dsafksjkfgksdjflk::', hsw.Zw_drape(Z,hw,i,j), hsw.Zw_drape(Z,hw,ir,jr), rat * hsw.Zw_drape(Z,hw,ir,jr) + (1 - rat) *  hsw.Zw_drape(Z,hw,i,j))


	# Traversing nodes
	for i,j in Z:
		
		# Updating local discharge to new time step
		QwA[i,j] = QwB[i,j]
		
		# Updating flow depth (cannot be < 0)
		# hw[i,j]  =  ti.math.clamp(
		# 					 hw[i,j] + (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy) ,
		# 					 # TODO try to keep track of the receivers and to force no inversion
		# 				constrains[i,j,0], # min
		# 				# constrains[i,j,1]  # max
		# 				1e6  # max
		# 			)
		# hw[i,j] = ti.math.max(0.,hw[i,j] + (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy) )

		if(QwA[i,j] <= 0 or abs(1 - (QwC[i,j]/QwA[i,j])) < threshold ):
			continue

		hw[i,j] = hw[i,j] + (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy) 

		# if(constrains[i,j,0] == 1e6):
		# 	constrains[i,j,0] = Z[i,j]
		# if(constrains[i,j,1] == 1e6):
		# 	constrains[i,j,1] = Z[i,j]

		# if(constrains[i,j,0] == constrains[i,j,1]):
		# 	print("HAPPENS")
		
		constrains[i,j,0] -= Z[i,j]
		constrains[i,j,1] -= Z[i,j]


		
		if(gridfuncs.can_out(i,j,BCs) == False):
			if(hw[i,j] < constrains[i,j,0]):
				hw[i,j] = constrains[i,j,0]
			elif(hw[i,j] > constrains[i,j,1]):
				hw[i,j] = constrains[i,j,1]
	

	# # DEBUG DEBUG DEBUG
	# # Traversing nodes
	# Nhap = 0
	# NNhap = 0
	# for i,j in Z:
	# 	if(D4dir[i,j] == 5 or gridfuncs.can_out(i,j,BCs)):
	# 		continue

	# 	ir,jr = gridfuncs.neighbours(i,j,D4dir[i,j], BCs)

	# 	if(hsw.Zw_drape(Z,hw,i,j) == hsw.Zw_drape(Z,hw,ir,jr) and (ir != i or jr != j)):
	# 		print('gabul', hsw.Zw_drape(Z,hw,i,j) ,'vs',hsw.Zw_drape(Z,hw,ir,jr), 'constrains node were', constrains[i,j,0] + Z[i,j], constrains[i,j,1] + Z[i,j], 'and rec', constrains[ir,jr,0] + Z[ir,jr], constrains[ir,jr,1] + Z[ir,jr]  )

	# # 	if( constrains[i,j,0] + Z[i,j] ==  constrains[ir,jr,0] + Z[ir,jr] or constrains[i,j,0] + Z[i,j] ==  constrains[ir,jr,1] + Z[ir,jr]  or constrains[i,j,1] + Z[i,j] ==  constrains[ir,jr,0] + Z[ir,jr] or constrains[i,j,1] + Z[i,j] ==  constrains[ir,jr,1] + Z[ir,jr]):
	# # 		Nhap +=1
	# # 	else:
	# # 		NNhap +=1
	# # print('error:', Nhap/(NNhap + Nhap))


@ti.kernel
def _compute_hw_drape_CFL(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), D4dir:ti.template(), constrains:ti.template(), BCs:ti.template(), alpha:ti.f32 ):
	'''
	EXPERIMENTAL
	Compute flow depth by div.Q and update discharge to t+1.
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- constrains: a 3D field of minimum [i,j,0] and maximum [i,j,1] Zw possible for every nodes
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, update flow depth in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	

	for i,j in Z:

		constrains[i,j,0] = -1e6
		constrains[i,j,1] = 1e6
		D4dir[i,j] = 5

	rat = 0.45

	for i,j in Z:

		tZw = hsw.Zw_drape(Z,hw,i,j) # + (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy)
		if gridfuncs.is_active(i,j,BCs) == False:
			continue

		kmin = 5
		Zdkmin = tZw

		# Traversing Neighbours
		for k in range(4):

			# Getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			if(ir == -1):
				continue

			# tSwr = hsw.Zw_drape(Z,hw,ir,jr) + (QwA[ir,jr] - QwC[ir,jr]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy)
			tZwr = hsw.Zw_drape(Z,hw,ir,jr)

			# if(tZwr == tZw):
			# 	print('HAPPENS:',i,j,'vs',ir,jr, tZwr,tZw)

			if(tZwr < Zdkmin):
				Zdkmin = tZwr
				kmin = k

		if kmin==5:
			continue

		ir,jr = gridfuncs.neighbours(i,j,kmin, BCs)


		D4dir[i,j] = kmin

		# rat = 0.45 + ti.random() * 0.04
		# rat = 1.

		constrains[i,j,0] = (rat * (hsw.Zw_drape(Z,hw,ir,jr)) + (1 - rat) * (tZw))
		# constrains[i,j,0] = hsw.Zw_drape(Z,hw,ir,jr)


		ti.atomic_min(constrains[ir,jr,1] , ((1-rat) * (hsw.Zw_drape(Z,hw,ir,jr)) + rat * (tZw)))

		# ti.atomic_min(constrains[ir,jr,1] , tZw)

		# constrains[i,j,0] -= Z[i,j]
		# constrains[i,j,1] -= Z[i,j]

	# for i,j in Z:
	# 	if(constrains[i,j,0] == constrains[i,j,1]):
	# 		ir,jr = gridfuncs.neighbours(i,j,D4dir[i,j], BCs)
	# 		print('dsafksjkfgksdjflk::', hsw.Zw_drape(Z,hw,i,j), hsw.Zw_drape(Z,hw,ir,jr), rat * hsw.Zw_drape(Z,hw,ir,jr) + (1 - rat) *  hsw.Zw_drape(Z,hw,i,j))


	# Traversing nodes
	for i,j in Z:
		
		# Updating local discharge to new time step
		QwA[i,j] = QwB[i,j]
		
		# Updating flow depth (cannot be < 0)
		# hw[i,j]  =  ti.math.clamp(
		# 					 hw[i,j] + (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy) ,
		# 					 # TODO try to keep track of the receivers and to force no inversion
		# 				constrains[i,j,0], # min
		# 				# constrains[i,j,1]  # max
		# 				1e6  # max
		# 			)
		# hw[i,j] = ti.math.max(0.,hw[i,j] + (QwA[i,j] - QwC[i,j]) * PARAMHYDRO.dt_hydro/(GRID.dx*GRID.dy) )
		tdt = ti.math.max(PARAMHYDRO.dt_hydro, alpha *GRID.dx/(QwA[i,j]/(ti.math.max(1e-4, hw[i,j]))) )		
		hw[i,j] = hw[i,j] + (QwA[i,j] - QwC[i,j]) * tdt/(GRID.dx*GRID.dy) 

		# if(constrains[i,j,0] == 1e6):
		# 	constrains[i,j,0] = Z[i,j]
		# if(constrains[i,j,1] == 1e6):
		# 	constrains[i,j,1] = Z[i,j]

		# if(constrains[i,j,0] == constrains[i,j,1]):
		# 	print("HAPPENS")
		
		constrains[i,j,0] -= Z[i,j]
		constrains[i,j,1] -= Z[i,j]


		
		if(gridfuncs.can_out(i,j,BCs) == False):
			if(hw[i,j] < constrains[i,j,0]):
				hw[i,j] = constrains[i,j,0]
			elif(hw[i,j] > constrains[i,j,1]):
				hw[i,j] = constrains[i,j,1]
	

	# # DEBUG DEBUG DEBUG
	# # Traversing nodes
	# Nhap = 0
	# NNhap = 0
	# for i,j in Z:
	# 	if(D4dir[i,j] == 5 or gridfuncs.can_out(i,j,BCs)):
	# 		continue

	# 	ir,jr = gridfuncs.neighbours(i,j,D4dir[i,j], BCs)

	# 	if(hsw.Zw_drape(Z,hw,i,j) == hsw.Zw_drape(Z,hw,ir,jr) and (ir != i or jr != j)):
	# 		print('gabul', hsw.Zw_drape(Z,hw,i,j) ,'vs',hsw.Zw_drape(Z,hw,ir,jr), 'constrains node were', constrains[i,j,0] + Z[i,j], constrains[i,j,1] + Z[i,j], 'and rec', constrains[ir,jr,0] + Z[ir,jr], constrains[ir,jr,1] + Z[ir,jr]  )

	# # 	if( constrains[i,j,0] + Z[i,j] ==  constrains[ir,jr,0] + Z[ir,jr] or constrains[i,j,0] + Z[i,j] ==  constrains[ir,jr,1] + Z[ir,jr]  or constrains[i,j,1] + Z[i,j] ==  constrains[ir,jr,0] + Z[ir,jr] or constrains[i,j,1] + Z[i,j] ==  constrains[ir,jr,1] + Z[ir,jr]):
	# # 		Nhap +=1
	# # 	else:
	# # 		NNhap +=1
	# # print('error:', Nhap/(NNhap + Nhap))



@ti.kernel
def _compute_link_based(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''

	for i,j in Z:
		QwB[i,j] = 0.
		QwC[i,j] = 0.
		# QwD[i,j] = 0.


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue


		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):


			while(True):
				# Traversing Neighbours
				for k in range(4):

					# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
					ir,jr = gridfuncs.neighbours(i,j,k, BCs)

					# if not a neighbours, by convention is < 0 and I pass
					if(ir == -1):
						continue

					if(gridfuncs.can_receive(ir,jr, BCs) == False):
						continue

					# Local hydraulic slope
					tS = hsw.Sw(Z,hw,i,j,ir,jr)

					# If < 0, neighbour is a donor and I am not interested
					if(tS <= 0):
						continue

					# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
					if(k == 0 or k == 3):
						if(tS > SSy):
							SSy = tS
					else:
						if(tS > SSx):
							SSx = tS

					# Registering local slope
					Sws[k] = tS
					# Summing it to global
					sumSw += tS
				
				if(sumSw > 0):
					break

				hw[i,j] += 1e-4

				# Done with processing this particular neighbour

			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			
			# Not sure I still need that
			if(gradSw == 0):
				continue

			# Transferring flow to neighbours
			for k in range(4):

				# local neighbours
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)
				
				# checking if neighbours
				if(ir == -1):
					continue

				if (Sws[k] <= 0 ):
					continue
				
				hw_eff = ti.max(hsw.Zw(Z,hw,i,j),hsw.Zw(Z,hw,ir,jr)) - ti.max(Z[i,j],Z[ir,jr])
				# hw_eff = ti.max(0,hw[i,j])
				tQw = GRID.dx * ti.math.pow(hw_eff,PARAMHYDRO.exponent_flow)/PARAMHYDRO.manning * Sws[k]/ti.math.sqrt(sumSw)
				QwC[i,j] += tQw

				# Transferring prop to the hydraulic slope
				ti.atomic_add(QwB[ir,jr], Sws[k]/sumSw * QwA[i,j])

		# Boundary case
		else:
			tSw = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val
			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * ti.math.sqrt(tSw)



@ti.kernel
def _archive_compute_Qw(Z:ti.template(), hw:ti.template(), QwA:ti.template(), QwB:ti.template(), QwC:ti.template(), BCs:ti.template(), flowdir:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''


	# Traversing each nodes
	for i,j in Z:

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.is_active(i,j,BCs) == False):
			continue

		# I'll store the hydraulic slope in this vector
		Sws = ti.math.vec4(0.,0.,0.,0.)

		# I'll need the sum of the hydraulic slopes in the positive directions
		sumSw = 0.

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SSx = 0.
		SSy = 0.

		# Safety check: gets incremented at each while iteration and manually breaks the loop if > 10k (avoid getting stuck in an infinite hole)
		lockcheck = 0

		if(gridfuncs.can_give(i,j,BCs) == False and gridfuncs.can_out(i,j,BCs) == False):
			continue

		thw = 0.

		# Tracking if I am in a local minima 
		LM = False

		# None boundary case
		if(gridfuncs.can_out(i,j,BCs) == False):
			# While I do not have external slope
			while(sumSw == 0.):
				
				# First incrementing the safety check
				lockcheck += 1

				# Traversing Neighbours
				for k in range(4):

					# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
					ir,jr = gridfuncs.neighbours(i,j,k, BCs)

					# if not a neighbours, by convention is < 0 and I pass
					if(ir == -1):
						continue

					if(gridfuncs.can_receive(ir,jr, BCs) == False):
						continue

					# Local hydraulic slope
					tS = hsw.Sw(Z,hw,i,j,ir,jr)

					# If < 0, neighbour is a donor and I am not interested
					if(tS <= 0):
						continue

					thw = ti.math.max(thw, ti.max(hsw.Zw(Z,hw,i,j),hsw.Zw(Z,hw,ir,jr)) - ti.max(Z[i,j],Z[ir,jr]))

					# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
					if(k == 0 or k == 3):
						if(tS > SSy):
							SSy = tS
					else:
						if(tS > SSx):
							SSx = tS

					# Registering local slope
					Sws[k] = tS
					# Summing it to global
					sumSw += tS

					# Done with processing this particular neighbour

				# Local minima management (cheap but works)
				## If I have no downward slope, I increase the elevation by a bit
				if(sumSw == 0.):
					# I am in a local minima
					LM = True
					hw[i,j] += 1e-4

				## And if I added like a metre and it did not slolve it, I stop for that node
				if(lockcheck > 10000 or (LM and PARAMHYDRO.use_original_dir_for_LM)):
					break

			# Calculating local norm for the gradient
			# The condition manages the boundary conditions
			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			
			# Not sure I still need that
			if(gradSw == 0):
				continue

			if(PARAMHYDRO.use_heffmax == False):
				thw = ti.math.max(0., hw[i,j])

			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(thw, PARAMHYDRO.exponent_flow) * sumSw/ti.math.sqrt(gradSw)

			# If I am in a local minima
			# either I propel the drainage area along a predefined railway to ensure its escape
			# or I just ignore it
			if(LM):
				# this option ensures the drainage of LM through the original rail
				if(PARAMHYDRO.use_original_dir_for_LM):
					# Flow dir == 5 is no flow
					if(flowdir[i,j] != 5):
						# That section ensures that a stochastic number of receivers are traversed to flush the local minima
						# And avoid ping-pong or localisation based biases
						ii,jj = i,j
						ir,jr = i,j
						first = 0
						# Receivers are poped out at least once, and then has a probability of 0.5 to continue
						while(flowdir[ir,jr] != 5 and (first<=PARAMHYDRO.LM_pathforcer)):
							ii,jj = ir,jr
							first += 1
							ir,jr = gridfuncs.neighbours(ii, jj, flowdir[ii,jj], BCs)
							# ti.atomic_add(QwB[ir,jr], QwA[i,j])
						ti.atomic_add(QwB[ir,jr], QwA[i,j])
				else:
					# In that case, I keep everything
					ti.atomic_add(QwB[i,j], QwA[i,j])

				continue

			# Transferring flow to neighbours
			for k in range(4):

				# local neighbours
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)
				
				# checking if neighbours
				if(ir == -1):
					continue
				
				# Transferring prop to the hydraulic slope
				ti.atomic_add(QwB[ir,jr], Sws[k]/sumSw * QwA[i,j])

		# Boundary case
		else:
			tSw = ti.max(hsw.Zw(Z,hw,i,j) -  PARAMHYDRO.hydro_slope_bc_val, 1e-6)/GRID.dx if PARAMHYDRO.hydro_slope_bc_mode == 0 else PARAMHYDRO.hydro_slope_bc_val
			# Calculating local discharge: manning's equations for velocity and u*h*W to get Q
			QwC[i,j] = GRID.dx/PARAMHYDRO.manning * ti.math.pow(ti.max(0.,hw[i,j]), PARAMHYDRO.exponent_flow) * ti.math.sqrt(tSw)




###########################################################
# Other sets of functions to help wiith hydrodynamcis
###########################################################

@ti.kernel
def check_convergence(QwA : ti.template(), QwC : ti.template(), tolerance:ti.f32, converged:ti.template(), BCs:ti.template()):
	'''
	Warning, slow function-ish (to call every 100s of iterations is OK) that check the proportion of nodes that have reached convergence.
	Only valid for steady flow assumption.
	Computes the ratio between Qw_out and Qw_in and save the proportion of nodes that have reached convergence within a tolerance factor.

	Arguments:
		- QwA: Discharge input to every cell (calculated by the compute_Qw function)
		- QwC: Discharge output to every cell (calculated by the compute_Qw function)
		- tolerance: determines if a node has converged if its tolerance > |1 - ratio|
		- converged: the convergence rate = number of nodes converged / total number of (active) nodes
		- BCs: the field of boundary conditions
		

	'''
	# Final count of converged
	count = 0
	# Total number of active nodes
	tot = 0
	# main loop
	for i,j in QwA:
		# Ignoring points without QwA
		if(QwA[i,j] > 0):
			# Is active: incrementing
			tot += 1
			# Ration Qwout / Qwin
			rat = QwC[i,j]/QwA[i,j]
			if(rat >= 1-tolerance and rat < (1 + tolerance)):
				count += 1

	# The final scalar result
	converged[None] = count/tot



########################################################################
########################################################################
############### EXPOSED API ############################################
########################################################################
########################################################################

# Note that the input functions are already exposed

compute_hw = None
compute_Qw = None


def set_hydro_CC():
	'''
	Expose the right API function of the comnpile time parameters in PARAMHYDRO.
	Need to be called after setting hte different parameters in the singleton PARAMHYDRO
	Returns:
		- Nothing, update global function refs
	Authors:
		- B.G. (last modification 03/05/2024)
	'''	


	# # fectch neighbours placeholder
	global compute_hw
	global compute_Qw

	
	# Feed it
	if(PARAMHYDRO.flowmode == FlowMode.static_incremental):
		compute_hw = _compute_hw
		compute_Qw = _compute_Qw
	elif(PARAMHYDRO.flowmode == FlowMode.static_drape):
		compute_hw = _compute_hw_drape
		compute_Qw = _compute_Qw_drape
	elif(PARAMHYDRO.flowmode == FlowMode.static_link):
		compute_hw = _compute_hw
		compute_Qw = _compute_link_based
	else:
		raise NotImplementedError('FLOWMODEW Not implemented yet')



'''
Sets of functions to help with the GUI made with niceGUI
'''

import plotly.graph_objects as go
import cmcrameri.cm as cmc
from nicegui import ui
import numpy as np
from PIL import Image

def cmap2plotly(cmap = cmc.glasgow):
	'''
	Convert a matplotlib-type colormap into a plotly-compatible format
	
	Arguments:
		- cmap: the Matplotlib-like colormap (can be a cmcrameri one)
	
	Returns:
		- the plotly-compatible colormap
	
	Authors:
		- B.G. (last modifications: 06/2024)
	'''

	# Dummy structure for the RGB
	cmap_colors = cmap(np.linspace(0, 1, 256))

	# Colorscale is the plotly colormap
	colorscale = [[i / 255, f'rgb({r*255},{g*255},{b*255})'] for i, (r, g, b, _) in enumerate(cmap_colors)]

	return colorscale
	


def notify_WIP():
	ui.notify("Work In Progress...")

def quit_app():
	ui.run_javascript('window.close()')


'''
Riverdale's mirror for the numba engine (steenbock) convention for the nth neighbouring:

B.G - 07/2024 - Acigné

'''

import numba as nb
import numpy as np
from enum import Enum
import scabbard.utils as scaut 


#################################################################################################
############################## Customs Boundaries ###############################################
#################################################################################################

'''
Reminder, I am using the DAGGER convention
// Cannot flow at all = nodata
NO_FLOW = 0,

// Internal Node (can flow in every directions)
FLOW = 1,

// Internal Node (can flow in every directions) BUT neighbours a special flow
// condition and may need specific care
FLOW_BUT = 2,

// flow can out there but can also flow to downstream neighbours
CAN_OUT = 3,

// flow can only out from this cell
OUT = 4,

// Not only flow HAS to out there: neighbouring flows will be drained there no
// matter what
FORCE_OUT = 5,

// Flows through the cell is possible, but the cell CANNOT out fluxes from
// this boundary (reserved to model edges, internal boundaries wont give to
// nodata anyway)
CANNOT_OUT = 6,

// Flow can only flow to potential receivers
IN = 7,

// Forced INFLOW: flow will flow to all neighbours (except other FORCE_IN)
FORCE_IN = 8,

// periodic border
PERIODIC_BORDER = 9
'''

@nb.njit()
def _check_top_row_customs_D4(i:int, j:int, k:int, BCs, valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the top row
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first row
	if(i == 0):
		if(k == 0):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_top_row_customs_D4_flat(i:int, k:int, BCs, valid:bool, nx:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the top row
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first row
	if(i < nx):
		if(k == 0):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_top_row_customs_D8(i:int, j:int, k:int, BCs, valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the top row
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first row
	if(i == 0):
		if(k == 0 or k ==1 or k==2):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_top_row_customs_D8_flat(i:int, k:int, BCs, valid:bool, nx:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the top row
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first row
	if(i < nx):
		if(k == 0 or k ==1 or k==2):
			valid = False
	# Done
	return valid


@nb.njit()
def _check_leftest_col_customs_D4(i:int, j:int, k:int, BCs, valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the leftest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first col
	if(j == 0):
		if(k==1):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_leftest_col_customs_D4_flat(i:int, k:int, BCs, valid:bool, nx:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the leftest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first col
	if( (i%nx) == 0):
		if(k==1):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_leftest_col_customs_D8(i:int, j:int, k:int, BCs, valid:bool):
	'''
	Internal function to check if neighbouring is possible for nodes at the leftest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first col
	if(j == 0):
		if(k==0 or k == 3 or k == 5):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_leftest_col_customs_D8_flat(i:int, k:int, BCs, valid:bool, nx:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the leftest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first col
	if((i % nx) == 0):
		if(k==0 or k == 3 or k == 5):
			valid = False
	# Done
	return valid


@nb.njit()
def _check_rightest_col_customs_D4(i:int, j:int, k:int, BCs, valid:bool, nx:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the rightest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first col
	if(j == nx-1):
		if(k==2):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_rightest_col_customs_D4_flat(i:int, k:int, BCs, valid:bool, nx:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the rightest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first col
	if(i%nx == nx-1):
		if(k==2):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_rightest_col_customs_D8(i:int, j:int, k:int, BCs, valid:bool, nx:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the rightest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first col
	if(j == nx-1):
		if(k==2 or k == 4 or k == 7):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_rightest_col_customs_D8_flat(i:int, k:int, BCs, valid:bool, nx:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the rightest column
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first col
	if(i%nx == nx-1):
		if(k==2 or k == 4 or k == 7):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_bottom_row_customs_D4(i:int, j:int, k:int, BCs, valid:bool, ny:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the bottom row
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first row
	if(i == ny-1):
		# Checking all the different cases: firs, last cols and the middle
		if(k == 3):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_bottom_row_customs_D4_flat(i:int, k:int, BCs, valid:bool, nx:int, ny:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the bottom row
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first row
	if(i >= nx*ny - nx):
		# Checking all the different cases: firs, last cols and the middle
		if(k == 3):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_bottom_row_customs_D8(i:int, j:int, k:int, BCs, valid:bool, ny:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the bottom row
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first row
	if(i == ny-1):
		# Checking all the different cases: firs, last cols and the middle
		if(k == 5 or k == 6 or k == 7):
			valid = False
	# Done
	return valid

@nb.njit()
def _check_bottom_row_customs_D8_flat(i:int, k:int, BCs, valid:bool, nx:int, ny:int):
	'''
	Internal function to check if neighbouring is possible for nodes at the bottom row
	Caution: this is optimised for neighbouring checks and ignores the top and bottom rows
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''
	# Only checking if it actually is on the first row
	if(i >= nx*ny - nx):
		# Checking all the different cases: firs, last cols and the middle
		if(k == 5 or k == 6 or k == 7):
			valid = False
	# Done
	return valid

@nb.njit()
def _cast_neighbour_customs_D4(i:int, j:int, k:int, valid:bool, BCs):
	'''
	Internal function that cast the neighbours to the right values in the case of normal boundary conditions
	Caution: this is optimised for neighbouring checks and should not be used on its own
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
		- valid: a boolean from previous checks 
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''

	# Preformat the output
	ir,jr = -1,-1

	# if the neighbouring operation is still valid after that:
	if(valid):
		if(k == 0):
			ir,jr = i-1, j
		if(k == 1):
			ir,jr = i, j-1
		if(k == 2):
			ir,jr = i, j+1
		if(k == 3):
			ir,jr = i+1, j

	if(BCs[i,j] == 0 or ir == -1):
		ir,jr = -1,-1
	elif(BCs[ir,jr] == 0):
		ir,jr = -1,-1
		

	return ir, jr

@nb.njit()
def _cast_neighbour_customs_D4_flat(i:int, k:int, valid:bool, BCs, nx:int):
	'''
	Internal function that cast the neighbours to the right values in the case of normal boundary conditions
	Caution: this is optimised for neighbouring checks and should not be used on its own
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
		- valid: a boolean from previous checks 
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''

	# Preformat the output
	ir = -1

	# if the neighbouring operation is still valid after that:
	if(valid):
		if(k == 0):
			ir = np.int64(i-nx)
		if(k == 1):
			ir = np.int64(i-1)
		if(k == 2):
			ir = np.int64(i+1)
		if(k == 3):
			ir = np.int64(i+nx)

	if(BCs[i] == 0 or ir == -1):
		ir = np.int64(-1)
	elif(BCs[ir] == 0):
		ir = np.int64(-1)

	return ir


@nb.njit()
def _cast_neighbour_customs_D8(i:int, j:int, k:int, valid:bool, BCs):
	'''
	Internal function that cast the neighbours to the right values in the case of normal boundary conditions
	Caution: this is optimised for neighbouring checks and should not be used on its own
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
		- valid: a boolean from previous checks 
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''

	# Preformat the output
	ir,jr = -1,-1

	# if the neighbouring operation is still valid after that:
	if(valid):
		if(k == 0):
			ir,jr = i-1, j-1
		elif(k == 1):
			ir,jr = i-1, j
		elif(k == 2):
			ir,jr = i-1, j+1
		elif(k == 3):
			ir,jr = i, j-1
		elif(k == 4):
			ir,jr = i, j+1
		elif(k == 5):
			ir,jr = i+1, j-1
		elif(k == 6):
			ir,jr = i+1, j
		elif(k == 7):
			ir,jr = i+1, j+1

	if(BCs[i,j] == 0 or ir == -1):
		ir,jr = -1,-1
	elif(BCs[ir,jr] == 0):
		ir,jr = -1,-1
		
	return ir, jr


@nb.njit()
def _cast_neighbour_customs_D8_flat(i:int, k:int, valid:bool, BCs, nx:int):
	'''
	Internal function that cast the neighbours to the right values in the case of normal boundary conditions
	Caution: this is optimised for neighbouring checks and should not be used on its own
	Arguments:
		- i: Row index
		- j: column index
		- k: neighbour number (See top of this module for explanations)
		- valid: a boolean from previous checks 
	Returns:
		- a boolean: True = neighbour is valid, False: not a neighbour
	Authors:
		- B.G. (last modification 02/05/2024)
	'''

	# Preformat the output
	ir:nb.int64 = -1
	# if the neighbouring operation is still valid after that:
	if(valid):
		if(k == 0):
			ir = np.int64(i-nx-1)
		elif(k == 1):
			ir = np.int64(i-nx)
		elif(k == 2):
			ir = np.int64(i-nx+1)
		elif(k == 3):
			ir = np.int64(i-1)
		elif(k == 4):
			ir = np.int64(i+1)
		elif(k == 5):
			ir = np.int64(i+nx-1)
		elif(k == 6):
			ir = np.int64(i+nx)
		elif(k == 7):
			ir = np.int64(i+nx+1)

	if(BCs[i] == 0 or ir == -1):
		ir = np.int64(-1)
	elif(BCs[ir] == 0):
		ir = np.int64(-1)
		
	return ir

@nb.njit()
def neighbours_D4(i:int, j:int, k:int, BCs, nx:int, ny:int):
	'''
	GPU function returning the neighbours of a given pixel
	Arguments:\
		- i,j are the row and col indices
		- k is the nth neighbour (4 in D4) following riverdale's convention (see top of this module)
		- BCs: boundary conditions code. Note that for this function it does not do anything and won't be used but it complies to the standard
	Returns:
		- (-1,-1) if hte neighbour is not valid (e.g. normal boundaries at the left border has no left neighbour k=1)
		- the indices of the row/col of the neighbours
	Authors:
		- B.G. (last modification 02/05/2024)
	TODO:
		- adding the Periodic boundary management in the checks
	'''

	# I first assume this mneighbour is valid
	valid = True

	# Breaking down the checks
	valid = _check_top_row_customs_D4(i,j,k,BCs,valid)
	valid = _check_leftest_col_customs_D4(i,j,k,BCs,valid)
	valid = _check_rightest_col_customs_D4(i,j,k,BCs,valid,nx)
	valid = _check_bottom_row_customs_D4(i,j,k,BCs,valid,ny)

	# getting the actual neighbours
	return _cast_neighbour_customs_D4(i,j,k,valid,BCs)

@nb.njit()
def neighbours_D4_flat(i:int, k:int, BCs, nx:int, ny:int):
	'''
	GPU function returning the neighbours of a given pixel
	Arguments:\
		- i,j are the row and col indices
		- k is the nth neighbour (4 in D4) following riverdale's convention (see top of this module)
		- BCs: boundary conditions code. Note that for this function it does not do anything and won't be used but it complies to the standard
	Returns:
		- (-1,-1) if hte neighbour is not valid (e.g. normal boundaries at the left border has no left neighbour k=1)
		- the indices of the row/col of the neighbours
	Authors:
		- B.G. (last modification 02/05/2024)
	TODO:
		- adding the Periodic boundary management in the checks
	'''

	# I first assume this mneighbour is valid
	valid = True

	# Breaking down the checks
	valid = _check_top_row_customs_D4_flat(i,k,BCs,valid,nx)
	valid = _check_leftest_col_customs_D4_flat(i,k,BCs,valid,nx)
	valid = _check_rightest_col_customs_D4_flat(i,k,BCs,valid,nx)
	valid = _check_bottom_row_customs_D4_flat(i,k,BCs,valid,nx,ny)

	# getting the actual neighbours
	return _cast_neighbour_customs_D4_flat(i,k,valid,BCs,nx)


@nb.njit()
def neighbours_D8(i:int, j:int, k:int, BCs, nx:int, ny:int):
	'''
	GPU function returning the neighbours of a given pixel
	Arguments:\
		- i,j are the row and col indices
		- k is the nth neighbour (4 in D8) following riverdale's convention (see top of this module)
		- BCs: boundary conditions code. Note that for this function it does not do anything and won't be used but it complies to the standard
	Returns:
		- (-1,-1) if hte neighbour is not valid (e.g. normal boundaries at the left border has no left neighbour k=1)
		- the indices of the row/col of the neighbours
	Authors:
		- B.G. (last modification 02/05/2024)
	TODO:
		- adding the Periodic boundary management in the checks
	'''

	# I first assume this mneighbour is valid
	valid = True

	# Breaking down the checks
	valid = _check_top_row_customs_D8(i,j,k,BCs,valid)
	valid = _check_leftest_col_customs_D8(i,j,k,BCs,valid)
	valid = _check_rightest_col_customs_D8(i,j,k,BCs,valid,nx)
	valid = _check_bottom_row_customs_D8(i,j,k,BCs,valid,ny)

	# getting the actual neighbours
	return _cast_neighbour_customs_D8(i,j,k,valid,BCs)

@nb.njit()
def neighbours_D8_flat(i:int, k:int, BCs, nx:int, ny:int):
	'''
	GPU function returning the neighbours of a given pixel
	Arguments:\
		- i,j are the row and col indices
		- k is the nth neighbour (4 in D8) following riverdale's convention (see top of this module)
		- BCs: boundary conditions code. Note that for this function it does not do anything and won't be used but it complies to the standard
	Returns:
		- (-1,-1) if hte neighbour is not valid (e.g. normal boundaries at the left border has no left neighbour k=1)
		- the indices of the row/col of the neighbours
	Authors:
		- B.G. (last modification 02/05/2024)
	TODO:
		- adding the Periodic boundary management in the checks
	'''

	# I first assume this mneighbour is valid
	valid = True

	# Breaking down the checks
	valid = _check_top_row_customs_D8_flat(i,k,BCs,valid,nx)
	valid = _check_leftest_col_customs_D8_flat(i,k,BCs,valid,nx)
	valid = _check_rightest_col_customs_D8_flat(i,k,BCs,valid,nx)
	valid = _check_bottom_row_customs_D8_flat(i,k,BCs,valid,nx,ny)

	# getting the actual neighbours
	return _cast_neighbour_customs_D8_flat(i,k,valid,BCs,nx)






@nb.njit()
def can_receive(i:int, j:int, BCs):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
	'''
	valid = True
	if(BCs[i,j] == 6 or BCs[i,j] == 7 or BCs[i,j] == 8 or BCs[i,j] == 0):
		valid = False
	return valid

@nb.njit()
def can_give(i:int, j:int, BCs):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
		Authors:
		- B.G. (last modification 02/05/2024)
	'''
	valid = False
	if(BCs[i,j] == 1 or BCs[i,j] == 6 or BCs[i,j] == 7 or BCs[i,j] == 8 or BCs[i,j] == 9):
		valid = True
	return valid


@nb.njit()
def can_out(i:int, j:int, BCs):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
		Authors:
		- B.G. (last modification 02/05/2024)
	'''
	valid = False
	if(BCs[i,j] == 3 or BCs[i,j] == 4 or BCs[i,j] == 5):
		valid = True
	return valid


@nb.njit()
def can_receive_flat(i:int, BCs):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
	'''
	valid = True
	if(BCs[i] == 6 or BCs[i] == 7 or BCs[i] == 8 or BCs[i] == 0):
		valid = False
	return valid

@nb.njit()
def can_give_flat(i:int, BCs):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
		Authors:
		- B.G. (last modification 02/05/2024)
	'''
	valid = False
	if(BCs[i] == 1 or BCs[i] == 6 or BCs[i] == 7 or BCs[i] == 8 or BCs[i] == 9):
		valid = True
	return valid


@nb.njit()
def can_out_flat(i:int, BCs):
	'''
		Standard complying function for the normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True, all the nodes can receive in the normal boundary conditions
		Authors:
		- B.G. (last modification 02/05/2024)
	'''
	valid = False
	if(BCs[i] == 3 or BCs[i] == 4 or BCs[i] == 5):
		valid = True
	return valid


@nb.njit()
def is_active_flat(i:int, BCs):
	'''
		Quick utility function determining if a node is active or not for normal boundaries
		Arguments:
			- i: the row index
			- j: the column index
			- BCs: a dummy field to keep the standard consistent
		Returns:
			- True if the node is active
			- False if inactive (i.e in this case outs)
		Authors:
		- B.G. (last modification 02/05/2024)
	'''
	valid = True
	if(BCs[i] == 0): #) or _can_out_customs(i,j,BCs) or _can_receive_customs(i,j,BCs) == False):
		valid = False
	return valid



########################################################################
########################################################################
############### GENERIC  FUNCTIONS #####################################
########################################################################
########################################################################


@nb.njit()
def oppk_D4(k):
	'''
	Returns the opposite neighbour code, e.g. if supplied with 1 (left neighbour), returns 2 (right neighbour)
	Useful to check if a k neighbour points toward a cell
	
	Arguments:
		k: the neihgbour code

	returns:
		The opposite neighbour code

	Authors:
		- B.G. (last modifications: 06/2024)
	'''
	return 3 if k == 0 else (2 if k == 1 else (1 if k == 2 else (0 if k == 3 else 5)))

@nb.njit()
def oppk_D8(k):
	'''
	Returns the opposite neighbour code, e.g. if supplied with 1 (left neighbour), returns 2 (right neighbour)
	Useful to check if a k neighbour points toward a cell
	
	Arguments:
		k: the neihgbour code

	returns:
		The opposite neighbour code

	Authors:
		- B.G. (last modifications: 06/2024)
	'''
	return 7 if k == 0 else (6 if k == 1 else (5 if k == 2 else (4 if k == 3 else (3 if k == 4 else (2 if k == 5 else (1 if k==6 else (0))))) ) )


@nb.njit()
def dx_from_k_D4(dx, k):
	'''
	Gets the distance to the neighbours
	'''
	return dx

@nb.njit()
def dx_from_k_D8(dx, k):
	'''
	Gets the distance to the neighbours
	'''
	return dx if (k == 1 or k == 3 or k == 4 or k == 6) else 1.41421356237*dx

@nb.njit()
def dy_from_k_D4(dx, k):
	'''
	Gets the distance to the neighbours
	'''
	return dx

@nb.njit()
def dy_from_k_D8(dx, k):
	'''
	Gets the distance to the neighbours
	'''
	return dx if (k == 1 or k == 3 or k == 4 or k == 6) == False else 1.41421356237*dx



'''
Helper functions to offset the actual code from the nice_haguid and let it focus on layout
'''
import scabbard as scb
from nicegui import ui
import os
import numpy as np
from scabbard.riverdale.rd_params import param_from_dem
from scabbard.riverdale.rd_env import create_from_params, load_riverdale
import taichi as ti
import scabbard as scb
import matplotlib.pyplot as plt
import numpy as np
import time
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.riverdale.rd_hydrodynamics as hyd
import scabbard.riverdale.rd_hydrometrics as rdta
import scabbard.riverdale.rd_LM as lm
import scabbard.riverdale.rd_drainage_area as rda
from scabbard.riverdale.rd_hillshading import hillshading
import plotly.graph_objects as go
import cmcrameri.cm as cmc
import scabbard.nice_utils as nut
import sys


def _update_clim(stuff):
	stuff['main_figure'].update_traces(
		zmin=stuff['model']['range']['min'],
		zmax=stuff['model']['range']['max'],
		selector=dict(name='datamap')
	)

	stuff['ui']['plot'].update()


"""
This class deals with loading raster informations
Authors: B.G.
"""
import numpy as np
import rasterio as rio
from rasterio.transform import from_bounds
import dagger as dag
from rasterio.transform import from_origin


def load_raster(fname):
	"""
	Load a raster array with different options. It uses rasterio that itself uses gdal.
	Arguments:
		fname (str): raster to load (path+file_name+format)
	Returns:
		A python dictionnary containing the following "key" -> val:
			"res" -> Resolution of the DEM
			"ncols" -> number of columns
			"nrows" -> number of rows
			"x_min" -> well x minimum
			"y_min" -> and y minimum
			"x_max" -> and x maximum
			"y_max" -> and x maximum
			"extent" -> extent combined in order to match matplotlib
			"array" -> numpy 2D array containing the data
			"crs" -> The crs string (geolocalisation)
			"nodata" -> list of nodata values
	Authors:
		B.G.
	Date:
		23/02/2019
	"""

	# Loading the raster with rasterio
	this_raster = rio.open(fname)

	# Initialising a dictionary containing the raster info for output
	out = {}
	# I am padding a no_data contour
	gt = this_raster.res
	out['dx'] = gt[0]
	out['dy'] = gt[1]
	out["nx"] = this_raster.width
	out["ny"] = this_raster.height
	out["x_min"] = this_raster.bounds[0]
	out["y_min"] = this_raster.bounds[1]
	out["x_max"] = this_raster.bounds[2]
	out["y_max"] = this_raster.bounds[3]
	corr = out['dx'] + out['dy']
	out["extent"] = [out["x_min"],out["x_max"]-corr,out["y_min"],out["y_max"]-corr]
	out["array"] = this_raster.read(1).astype(np.float64)
	try:
		out['crs'] = this_raster.crs['init']
	except (TypeError, KeyError) as e:
		out['crs'] = u'epsg:32601'
	out['nodata'] = this_raster.nodatavals

	
	
	# pixelSizeY =-gt[4]
	# (left=358485.0, bottom=4028985.0, right=590415.0, top=4265115.0)

	return out




def raster2graphcon(file_name):
	"""
	Ingest a raster with rasterio adn load a connector and graph from it
	"""
		
	# Loading DEM data with rasterio
	dem = load_raster(file_name)

	connector = dag.D8N(dem["nx"], dem["ny"], dem["dx"], dem["dy"], dem["x_min"], dem["y_min"])
	graph = dag.graph(connector)
	
	
	
	return connector, graph, dem

def raster2con(file_name):
	"""
	Ingest a raster with rasterio adn load a connector and graph from it
	"""
		
	# Loading DEM data with rasterio
	dem = load_raster(file_name)

	connector = dag.D8N(dem["nx"], dem["ny"], dem["dx"], dem["dy"], dem["x_min"], dem["y_min"])
	
	return connector, dem


# import rasterio
def save_raster(file_name, array, x_min, x_max, y_min, y_max, dx, dy, crs = None):
	# Define file name and file mode
	# file_name = "output.tif"
	file_mode = "w+"

	# Specify dimensions of the array
	height, width = array.shape
	count = 1  # Number of bands in the array
	dtype = array.dtype

	# Specify transformation parameters
	# x_min = 0  # X-coordinate of the top-left corner
	# y_max = 0  # Y-coordinate of the top-left corner
	x_res = dx  # Pixel size in the x-direction
	y_res = dy  # Pixel size in the y-direction
	transform = from_origin(x_min, y_max, dx, dy)

	if(crs is None):
		crs = "EPSG:35653"

	# Create output GeoTIFF file
	with rio.open(file_name, file_mode, driver='GTiff', height=height, width=width, dtype=dtype, count=count, transform=transform, nodata = -9999, crs = crs) as dst:
	    dst.write(array, 1)  # Write the array to the GeoTIFF file as the first band (band index is 1-based)



"""
Main file running the Fastflood model on the python side
"""

import dagger as dag
import numpy as np
import matplotlib.pyplot as plt
from scabbard import io
from scabbard import enumeration as en
import math
import warnings

class FastFlood(object):
	"""docstring for FastFlood"""
	def __init__(self):
		# To provide multiple constructors, I moved all the constructor to dedicated functions at the end of the page
		# One may want to load a file, start with a numpy array, generate random topo, ...
		pass



	def set_precipitations(self, precipitation):
		if(isinstance(precipitation, np.ndarray) == False):
			self.precipitation = np.full_like(self.topography,precipitation)
		else:
			self.precipitation = precipitation.ravel()
		self._.set_Qbase(self.precipitation * self.cellarea)



	def run(self, dt = 1e-3, method = en.Flooder.SFD_STATIC, experimental = False, **kwargs):

		if(isinstance(dt,np.ndarray)):
			self._.spatial_dt(dt.ravel())
		else:
			self._.set_dt(dt)


		if(self.monitor_dhw):
			self.dhw = self.get_hw()

		if(method == en.Flooder.SFD_STATIC):
			self._.run_SFD()

		elif(method == en.Flooder.MFD_STATIC):
			if(experimental == False):
				self._.run_MFD()
			else:
				self._.run_MFD_exp( kwargs["N_MFD_NOGRAPH"])


		elif method == en.Flooder.MFD_DYNAMIC:
			self._.run_MFD_dynamic()

		elif method == en.Flooder.CAESAR_LS:
			self._.caesar_lisflood()
		elif method == en.Flooder.CAESAR_LS_OMP:
			self._.caesar_lisflood_OMP()

		if(self.monitor_dhw):
			self.dhw = self.get_hw() - self.dhw 



	def get_hw(self):
		
		return self._.get_hw().reshape(self.rshp)

	def get_dhw(self):
		if(self.monitor_dhw == False):
			raise AttributeError("Cannot access to dhw if the option monitor_dhw is not active. ")
		return self.dhw.reshape(self.rshp)

	def get_Qwin(self):
		return self._.get_Qwin().reshape(self.rshp)

	def get_Qwout(self):
		return self._.get_Qwout().reshape(self.rshp)

	def get_HS(self):
		if(self.topology == en.Topology.D8):
			return dag.hillshade(self.connector, self._.get_topography()).reshape(self.rshp)
		else:
			warnings.warn("D4 hillshading not available, backporting to d8 (can take time)")
			tdag = dag.D8N(self.nx, self.ny, self.dx, self.dy, self.x_min, self.y_min)
			return dag.hillshade(tdag,self._.get_topography()).reshape(self.rshp)

	def get_topography(self):
		return self._.get_topography().reshape(self.rshp)





















def FF_from_file(file_name, precipitation = 1e-4, topology = en.Topology.D8):

	fastfloodobject = FastFlood()
		
	# Loading DEM data with rasterio
	dem = io.load_raster(file_name)
	
	# Storing topography
	fastfloodobject.topography = dem['array'].ravel()
	# model dimensions
	fastfloodobject.nx = dem['nx']
	fastfloodobject.ny = dem['ny']
	fastfloodobject.nxy = fastfloodobject.nx * fastfloodobject.ny
	fastfloodobject.dx = dem['dx']
	fastfloodobject.dy = dem['dy']
	fastfloodobject.cellarea = fastfloodobject.dx * fastfloodobject.dy
	fastfloodobject.dxy = math.sqrt(fastfloodobject.dx**2 + fastfloodobject.dy**2)
	fastfloodobject.x_min = dem["x_min"]
	fastfloodobject.x_max = dem["x_max"]
	fastfloodobject.y_min = dem["y_min"]
	fastfloodobject.y_max = dem["y_max"]
	fastfloodobject.extent = [fastfloodobject.x_min, fastfloodobject.x_max, fastfloodobject.y_min, fastfloodobject.y_max]

	# The connector
	fastfloodobject.connector = dag.D8N(dem["nx"], dem["ny"], dem["dx"], dem["dy"], dem["x_min"], dem["y_min"])
	fastfloodobject.DAG = dag.graph(fastfloodobject.connector)
	# fastfloodobject.DAG.init_graph(fastfloodobject.connector)
	fastfloodobject.DAG.set_LMR_method(dag.LMR.cordonnier_fill)
	fastfloodobject.rshp = (fastfloodobject.ny,fastfloodobject.nx)

	if(isinstance(precipitation, np.ndarray) == False):
		precipitation = np.full_like(fastfloodobject.topography,precipitation)

	fastfloodobject._ = dag.FF(fastfloodobject.DAG, fastfloodobject.connector, fastfloodobject.topography, precipitation.ravel())
	fastfloodobject.DAG.compute_graph(fastfloodobject.topography, True, True)
	fastfloodobject.topology = topology

	if(topology == en.Topology.D8):
		fastfloodobject._.set_topological_number(0.5)
	elif(topology == en.Topology.D4):
		fastfloodobject._.set_topological_number(1)


	# Monitoring options
	fastfloodobject.monitor_dhw = False
	fastfloodobject.dhw = np.zeros_like(fastfloodobject.topography)

	return fastfloodobject




def FF_from_array(nx,ny,dx,dy,xmin,ymin, array, topology = en.Topology.D8, precipitation = 1e-4, BCs = None):

	fastfloodobject = FastFlood()
		
	# Storing topography
	fastfloodobject.topography = array.ravel()
	# model dimensions
	fastfloodobject.nx = nx
	fastfloodobject.ny = ny
	fastfloodobject.nxy = fastfloodobject.nx * fastfloodobject.ny
	fastfloodobject.dx = dx
	fastfloodobject.dy = dy
	fastfloodobject.cellarea = fastfloodobject.dx * fastfloodobject.dy
	fastfloodobject.dxy = math.sqrt(fastfloodobject.dx**2 + fastfloodobject.dy**2)
	fastfloodobject.x_min = xmin
	fastfloodobject.x_max = xmin + (nx + 1) * dx
	fastfloodobject.y_min = ymin
	fastfloodobject.y_max = ymin + (ny + 1) * dy
	fastfloodobject.extent = [fastfloodobject.x_min, fastfloodobject.x_max, fastfloodobject.y_min, fastfloodobject.y_max]

	# The connector
	fastfloodobject.connector = dag.D8N(fastfloodobject.nx, fastfloodobject.ny, fastfloodobject.dx, fastfloodobject.dy, fastfloodobject.x_min, fastfloodobject.y_min) 
	
	if(BCs is not None):
		fastfloodobject.connector.set_custom_boundaries(BCs)

	# The graph
	fastfloodobject.DAG = dag.graph(fastfloodobject.connector)
	# fastfloodobject.DAG.init_graph(fastfloodobject.connector)
	fastfloodobject.DAG.set_LMR_method(dag.LMR.cordonnier_fill)
	fastfloodobject.rshp = (fastfloodobject.ny,fastfloodobject.nx)

	if(isinstance(precipitation, np.ndarray) == False):
		precipitation = np.full_like(fastfloodobject.topography,precipitation)

	fastfloodobject._ = dag.FF(fastfloodobject.DAG, fastfloodobject.connector, fastfloodobject.topography, precipitation.ravel())
	fastfloodobject.DAG.compute_graph(fastfloodobject.topography, True, True)
	fastfloodobject.topology = topology

	if(topology == en.Topology.D8):
		fastfloodobject._.set_topological_number(0.5)
	elif(topology == en.Topology.D4):
		fastfloodobject._.set_topological_number(1)


	# Monitoring options
	fastfloodobject.monitor_dhw = False
	fastfloodobject.dhw = np.zeros_like(fastfloodobject.topography)

	return fastfloodobject









		#end of class




'''
Contains generic widgets customisation for matplotlib qt backend
For example specific types of sliders (float, min max)
'''

import sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

# from PySide6 import QtWidgets, QtCore
from matplotlib.backends.qt_compat import QtWidgets, QtCore
import matplotlib
# Ensure using the Qt6Agg backend with PySide6
matplotlib.use('QtAgg')



class MainWindow(QtWidgets.QMainWindow):

	
	def __init__(self, title = "Main Window"):
		super(MainWindow, self).__init__()

		self.setWindowTitle(title)
		
		# Get the primary screen's geometry
		screen_geometry = QtWidgets.QApplication.primaryScreen().geometry()
		width, height = screen_geometry.width() * 0.9, screen_geometry.height() * 0.9

		# Calculate the position to center the window, then move it slightly down
		x = screen_geometry.width() * 0.05
		y = (screen_geometry.height() - height) / 2 + screen_geometry.height() * 0.02

		# Set the window geometry
		self.setGeometry(int(x), int(y), int(width), int(height))
		# grid layout is the easiest layout to organise everything by row and col
		self.centralWidget = QtWidgets.QWidget()
		self.grid_layout = QtWidgets.QGridLayout(self.centralWidget)
		# self.centralWidget.setLayout(self.grid_layout)
		self.setCentralWidget(self.centralWidget)
		self.widgets = {}
		self.layouts = {}
		self.layouts['main_grid'] = self.grid_layout


	def add_widget(self, ref, widget, row = None, col = None, rowmax = None, colmax = None, replace_ref = True, parentLayout = None):

		if ref in self.widgets.keys() and replace_ref:
			self.widgets[ref].deleteLater()
		# widget.setParentLayout(self)
		
		self.widgets[ref] = widget

		adada = self.layouts[parentLayout] if parentLayout else self.grid_layout


		if(rowmax):
			adada.addWidget(widget, row, col,rowmax, colmax)
		elif(row):
			adada.addWidget(widget, row, col)
		else:
			adada.addWidget(widget)


		adada.addStretch(1)

	def add_layout(self, ref, layout, row, col,rowmax = None, colmax = None, replace_ref = True, parentLayout = None, style = None):

		if ref in self.layouts.keys() and replace_ref:
			self.layouts[ref].deleteLater()
		# widget.setparentLayout(self)
		
		self.layouts[ref] = layout

		adada = self.layouts[parentLayout] if parentLayout else self.grid_layout

		twidget = QtWidgets.QWidget()
		twidget.setLayout(layout)
		if(style):
			twidget.setStyleSheet(style)


		if(rowmax):
			adada.addWidget(twidget, row, col,rowmax, colmax)
		else:
			adada.addWidget(twidget, row, col)






# end of file


'''
Numba scripts dedicated to local minimas
'''


import numpy as np
import numba as nb

@nb.njit()
def impose_downstream_minimum_elevation_decrease(Z, Sstack, Sreceivers, delta = 1e-4):

	for i in range(Z.shape[0]):
		node = Sstack[Z.shape[0] - 1 - i]
		rec = Sreceivers[node]

		if(node != rec):
			if(Z[rec] >= Z[node] - delta):
				Z[rec] = Z[node] - delta



'''
This is a first attempt to make a map automator.
Probably will just test different way here before making a better one
B.G
'''

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from scabbard import Dfig, Dax, RGridDax



def basemap(Rgrid, **kwargs):

	fig, ax = plt.subplots()
	if("figsize" in kwargs.keys()):
		fig.set_size_inches(*figsize)

	tDax = RGridDax(Rgrid, ax)
	tdfig = Dfig(fig, tDax)
	return tdfig






























#end of files



import numpy as np
import scabbard as scb
from scipy.ndimage import distance_transform_edt


def gaussian_fourier(
	grid: scb.raster.RegularRasterGrid, 
	in_place = False, 
	BCs = None, magnitude = 5):
	
	# masking
	mask = np.ones_like(grid.Z).astype(np.uint8) if (BCs is None) else (np.where(BCs == 0, 0, 1).astype(np.uint8))
	
	# Value to filter
	topography = grid.Z.copy()

	if(BCs is None) == False:
		# Compute the distance transform
		distance, indices = distance_transform_edt((mask == 0), return_indices=True)

		# Extract the nearest valid values using the indices
		topography = topography[tuple(indices)]
		

	# Perform the 2D Fourier Transform
	fourier_transform = np.fft.fft2(topography)
	fourier_transform_shifted = np.fft.fftshift(fourier_transform)

	# Define a low-pass filter: create a mask with ones in the low frequencies and zeros in the high frequencies
	rows, cols = topography.shape
	crow, ccol = rows // 2, cols // 2  # Center of the frequency domain

	# Create a Gaussian filter
	sigma = magnitude  # Adjust this value to control the level of smoothing (higher = more smoothing)
	y, x = np.ogrid[:rows, :cols]
	distance = np.sqrt((x - ccol)**2 + (y - crow)**2)
	gaussian_filter = np.exp(-(distance**2) / (2 * sigma**2))


	# Apply the Gaussian filter to the Fourier-transformed data
	filtered_fourier = fourier_transform_shifted * gaussian_filter

	# Perform the inverse Fourier Transform to reconstruct the smoothed topography
	inverse_shifted = np.fft.ifftshift(filtered_fourier)
	smoothed_topography = np.fft.ifft2(inverse_shifted).real.astype(grid.Z.dtype)  # Take the real part since the result might be complex

	smoothed_topography[mask == 0] = grid.Z[mask == 0]

	if in_place:
		grid.Z[:,:] = smoothed_topography[:,:]
	else:
		return grid.duplicate_with_other_data(smoothed_topography)


'''
Sets of function to compute drainage area metrics with riverdale
EXPERIMENTAL, no warranty it evens do what it is supposed to do yet

B.G. - 29/04/2024
'''

import taichi as ti
import numpy as np
from enum import Enum
import scabbard.utils as scaut 
from scabbard.riverdale.rd_grid import GRID
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.riverdale.rd_LM as lmfuncs
import scabbard.riverdale.rd_helper_surfw as hsw
import scabbard.riverdale.rd_utils as rut



@ti.kernel
def compute_Sreceivers_Zw(receivers:ti.template(), Z:ti.template(), hw:ti.template(), BCs:ti.template()):
	'''
	Compute the single flow receivers for each and every nodes
	Arguments:
		- receivers: 2D field holding the flat index of the steepest receivers
		- Z: topographic elevation
		- hw: flow depth
		- BCs: boundary condition codes
	Return:
		- Nothing by fill receivers in place
	Authors:
		- B.G. (last modifications: 06/2024)
	'''
	# Traversing each nodes
	for i,j in Z:

		# Assigning the receiver to itself by default
		receivers[i,j] = i * GRID.nx + j

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.is_active(i,j,BCs) == False):
			continue

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SS = 0.
		lowest_higher_Z = 0.
	
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			# Local hydraulic slope
			tS = hsw.Zw(Z,hw,i,j) - hsw.Zw(Z,hw,ir,jr)
			tS /= GRID.dx

			# If < 0, neighbour is a donor and I am not interested
			if(tS <= 0):
				continue

			# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
			if(tS > SS):
				receivers[i,j] = ir * GRID.nx + jr
				SS = tS

@ti.kernel
def compute_Sreceivers_Zw_rand(receivers:ti.template(), Z:ti.template(), hw:ti.template(), BCs:ti.template()):
	'''
	Compute the single flow receivers for each and every nodes
	This variant adds a stochastic component to the selection of the single flow receiver
	Arguments:
		- receivers: 2D field holding the flat index of the steepest receivers
		- Z: topographic elevation
		- hw: flow depth
		- BCs: boundary condition codes
	Return:
		- Nothing by fill receivers in place
	Authors:
		- B.G. (last modifications: 06/2024)
	'''
	# Traversing each nodes
	for i,j in Z:

		# Assigning the receiver to itself by default
		receivers[i,j] = i * GRID.nx + j

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.is_active(i,j,BCs) == False):
			continue

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SS = 0.
		lowest_higher_Z = 0.
	
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			# Local hydraulic slope
			tS = hsw.Zw(Z,hw,i,j) - hsw.Zw(Z,hw,ir,jr)
			tS /= GRID.dx
			# Randomiser yolo
			tS *= ti.random()

			# If < 0, neighbour is a donor and I am not interested
			if(tS <= 0):
				continue

			# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
			if(tS > SS):
				receivers[i,j] = ir * GRID.nx + jr
				SS = tS


@ti.kernel
def increment_DAD4(receivers:ti.template(), ptrrec:ti.template(), DA:ti.template()):
	'''
	Function incrementing the drainage area calculation by one step
	Based on a pointerto receiver method NOT OPTIMISED YET!
	Arguments:
		- receivers: 2D field of single flow receivers for each nodes
		- ptrrec: 2D field of pointer to the current receiver to atomic_add
		- DA: 2D field of drainage area
	Return:
		- nothing but edit receivers in place
	Authors:
		- B.G (last modifictations: 06/2024)
	'''

	# Main loop
	for i,j in receivers:
		
		# generating the flat index
		idx = i * GRID.nx + j
		
		# If outlet continue
		if(receivers[i,j] == idx):
			continue
		
		# row col indinces of the current pointer
		ip = ptrrec[i,j] // GRID.nx
		jp = ptrrec[i,j] %  GRID.nx

		# Flat index of the receiver to the current pointer
		newrec = receivers[ip,jp]

		# row col of the receivers of the pointer
		ni = newrec // GRID.nx
		nj = newrec %  GRID.nx

		# If is an outlet or local minima I stop here
		if(receivers[ni,nj] == newrec):
			continue

		# If it is not, I increment drainage area and ...
		ti.atomic_add(DA[ni,nj], GRID.dx * GRID.dy)
		# ... updating the pointer 
		ptrrec[i,j] = newrec

@ti.kernel
def increment_QWD4(receivers:ti.template(), ptrrec:ti.template(), QW:ti.template(), Pf:ti.template()):
	'''
	Function incrementing the drainage area calculation by one step
	Based on a pointerto receiver method NOT OPTIMISED YET!
	Arguments:
		- receivers: 2D field of single flow receivers for each nodes
		- ptrrec: 2D field of pointer to the current receiver to atomic_add
		- QW: 2D field of drainage area weighted by precipitations
		- Pf: 2D field of Precipitation inputs
	Return:
		- nothing but edit receivers in place
	Authors:
		- B.G (last modifictations: 06/2024)
	'''

	# Main loop
	for i,j in receivers:
		
		# generating the flat index
		idx = i * GRID.nx + j
		
		# If outlet continue
		if(receivers[i,j] == idx):
			continue
		
		# row col indinces of the current pointer
		ip = ptrrec[i,j] // GRID.nx
		jp = ptrrec[i,j] %  GRID.nx

		# Flat index of the receiver to the current pointer
		newrec = receivers[ip,jp]

		# row col of the receivers of the pointer
		ni = newrec // GRID.nx
		nj = newrec %  GRID.nx

		# If is an outlet or local minima I stop here
		if(receivers[ni,nj] == newrec):
			continue

		# If it is not, I increment drainage area and ...
		ti.atomic_add(QW[ni,nj], Pf[i,j])

		# print('adding', Pf[i,j])
		# ... updating the pointer 
		ptrrec[i,j] = newrec



def compute_drainage_area_D4(rd, fill = True, N = 'auto', random_rec = False, Precipitations = None):
	'''
	Compute drainage area in D4 direction on GPU. 
	Warning, this is optimised for flexibility to automate the computation fo DA for numerical analysis, not an optimised building block within a model (that's WIP)
	workflow: (optionally) fills the topo, then precalculates single flow receivers and finally transfer flow in a cascading bucekts fashion
	
	Arguments:
		- rd: the initialised RiverDale object
		- fill: if True, runs priority flood to get rid of local minimas
		- N: N iterations for the increment of DA (WIP to get that auto)
		- random_rec: if True, adds a stochastic component in the selection of receivers in the first step of computations
		- Precipitations: if 2D numpy array, accumulates P * drainage area
	Returns:
		- Numpy array of drainage area
	Authors:
		- B.G. (last modifications: 06/2024)
	'''	

	# Optional filling operation
	if(fill):
		lmfuncs.priority_flood(rd)

	# checking if precipitations are needed
	prec = False if Precipitations is None else True
	
	# Fetching fields for receivers and ptr to receivers	
	rec, ptrrec = rd.query_temporary_fields(2,dtype = ti.i32)
	# Fetching field for DA
	if(not prec):
		DA, = rd.query_temporary_fields(1, dtype = ti.f32) 
		# Initialising drainage area to local values
		DA.fill(GRID.dx * GRID.dy) 

	else:
		DA,Pfield = rd.query_temporary_fields(2, dtype = ti.f32) 
		Pfield.from_numpy(Precipitations * GRID.dx * GRID.dy)
		rut.A_equals_B(DA,Pfield)

	# Precomputing the receivers - with or without the stochastic component 
	compute_Sreceivers_Zw(rec, rd.Z, rd.hw, rd.BCs) if random_rec == False else compute_Sreceivers_Zw_rand(rec, rd.Z, rd.hw, rd.BCs)
	
	# initialising the ptr to receivers to the first receivers of each node
	rut.A_equals_B(ptrrec,rec)
	
	# If mode is 'auto', I set N to very high values
	if isinstance(N, str):
		N = 1000000
		auton = True
	else:
		auton = False
	
	#running N iterations (=/- debug logs)
	for i in range(N):
		
		if(i%1000 == 0 and i>0 and auton):
			cop = DA.to_numpy()
			increment_DAD4(rec,ptrrec,DA) if prec == False else increment_QWD4(rec, ptrrec, DA, Pfield)
			if(np.sum(np.abs(cop - DA.to_numpy())) == 0.):
				break
		else:
			increment_DAD4(rec,ptrrec,DA) if prec == False else increment_QWD4(rec, ptrrec, DA, Pfield)


	# Done, returning the results
	return DA.to_numpy()















@ti.kernel
def compute_D4_nolm(Z:ti.template(), D4dir:ti.template(), BCs:ti.template()):
	'''
	Experimental tests on drainage area calculations
	Do not use at the moment
	B.G.
	'''

	# Traversing each nodes
	for i,j in Z:

		D4dir[i,j] = ti.uint8(5)

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.is_active(i,j,BCs) == False):
			continue

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SS = 0.
		lowest_higher_Z = 0.
		checked = True

	
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			# Local hydraulic slope
			tS = Z[i,j] - Z[ir,jr]
			tS /= GRID.dx

			# If < 0, neighbour is a donor and I am not interested
			if(tS <= 0):
				if(Z[ir,jr] < lowest_higher_Z or lowest_higher_Z == 0.):
					lowest_higher_Z = Z[ir,jr]
				continue

			# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
			if(tS > SS):
				D4dir[i,j] = ti.uint8(k)
				SS = tS


			# Local minima management (cheap but works)
			## If I have no downward slope, I increase the elevation by a bit
			# if(SS == 0.):
			# 	if(checked):
			# 		checker[None] += 1
			# 		checked = False
			# 	Z[i,j] = max(lowest_higher_Z,Z[i,j]) + 1e-4

@ti.kernel
def compute_D4(Z:ti.template(), D4dir:ti.template(), BCs:ti.template(), checker:ti.template() ):
	'''
	Experimental tests on drainage area calculations
	Do not use at the moment
	B.G.
	'''

	# Traversing each nodes
	for i,j in Z:

		D4dir[i,j] = ti.uint8(5)

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.is_active(i,j,BCs) == False):
			continue

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SS = 0.
		lowest_higher_Z = 0.
		checked = True

		# While I do not have external slope
		while(SS == 0.):
			
			# Traversing Neighbours
			for k in range(4):
				# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
				ir,jr = gridfuncs.neighbours(i,j,k, BCs)

				# if not a neighbours, by convention is < 0 and I pass
				if(ir == -1):
					continue

				# Local hydraulic slope
				tS = Z[i,j] - Z[ir,jr]
				tS /= GRID.dx

				# If < 0, neighbour is a donor and I am not interested
				if(tS <= 0):
					if(Z[ir,jr] < lowest_higher_Z or lowest_higher_Z == 0.):
						lowest_higher_Z = Z[ir,jr]
					continue

				# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
				if(tS > SS):
					D4dir[i,j] = ti.uint8(k)
					SS = tS


			# Local minima management (cheap but works)
			## If I have no downward slope, I increase the elevation by a bit
			if(SS == 0.):
				if(checked):
					checker[None] += 1
					checked = False
				Z[i,j] = max(lowest_higher_Z,Z[i,j]) + 1e-4


	# if(globcheck > 0):
	# 	compute_D4(Z,D4dir,BCs)



















@ti.kernel
def step_DA_D4(Z:ti.template(), DA:ti.template(), temp:ti.template(), D4dir:ti.template(), BCs:ti.template() ):
	'''
	Compute and transfer QwA (in from t-1) into a temporary QwB (in for t).
	Also computes QwC (out at t) 
	Arguments:
		- Z: a 2D field of topographic elevation
		- hw: a 2D field of flow depth
		- QwA: a 2D field of discharge A (in)
		- QwB: a 2D field of discharge B (in t+1)
		- QwC: a 2D field of discharge C (out)
		- BCs: a 2D field of boundary conditions
	Returns:
		- Nothing, Caluclates disccharges in place
	Authors:
		- B.G. (last modification 03/05/2024)
	'''

	for i,j in Z:
		temp[i,j] = GRID.dx * GRID.dy
	
	for i,j in Z:
		if(gridfuncs.is_active(i,j,BCs)):
			if(D4dir[i,j] == 5):
				continue
			ir,jr = gridfuncs.neighbours(i,j,D4dir[i,j],BCs)
			if(ir>-1):
				ti.atomic_add(temp[ir,jr], DA[i,j])
	
	for i,j in Z:
		DA[i,j] = temp[i,j]
	

		



def launch_haGUId():

	import sys
	import numpy as np
	import matplotlib.pyplot as plt
	from matplotlib.backends.backend_qtagg import FigureCanvas
	from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

	# from PySide6 import QtWidgets, QtCore
	from matplotlib.backends.qt_compat import QtWidgets, QtCore
	import matplotlib
	from matplotlib.lines import Line2D

	import scabbard.phineas_helper as phelp
	from scabbard.gui_main_window import MainWindow
	from scabbard.gui_load_file import FileLoader
	from scabbard.gui_float_sliders import FloatSlider
	import scabbard.gui_maps_widgets as guimap
	import taichi as ti
	import scabbard as scb
	import numpy as np
	import matplotlib.pyplot as plt

	import threading

	# Create an Event object
	stop_event = threading.Event()

	
	fnamedem = f"/home/bgailleton/Desktop/code/FastFlood2.0/FastFlood2_Boris/graphflood/paper_scripts/data/green_river_{1}.tif"
	
	manning = 0.033
	env = scb.env_from_DEM(fnamedem)
	nx,ny = env.grid.nx, env.grid.ny
	dx,dy = env.grid.dx, env.grid.dy
	dt = 2e-3

	hmaxplot = 3

	ti.init(arch=ti.gpu)  # Initialize Taichi to use the CPU


	Z = ti.field(dtype=ti.f32, shape=(ny,nx))
	Z.from_numpy(env.grid.Z2D.astype(np.float32))
	hw = ti.field(dtype=ti.f32, shape=(ny,nx))

	QwA = ti.field(dtype=ti.f32, shape=(ny,nx))
	QwB = ti.field(dtype=ti.f32, shape=(ny,nx))
	QwC = ti.field(dtype=ti.f32, shape=(ny,nx))

	P = ti.field(dtype = ti.f32, shape = ())
	P[None] = np.float32(2e-5)







	@ti.kernel
	def init_field():
		for i, j in Z:
			hw[i,j] = 0

	@ti.kernel
	def stepinit():
		for i,j in Z:
			QwA[i,j] += P[None] * dx * dy
			QwB[i,j] = P[None] * dx * dy

	@ti.func
	def Zw(i,j) -> ti.f32:
		return Z[i,j] + hw[i,j]

	@ti.func
	def Sw(i,j, ir,jr)->ti.f32:
		return (Zw(i,j) - Zw(ir,jr))/dx

	@ti.func
	def Qw(i,j, tSw:ti.f32)->ti.f32:
		return dx/manning * ti.math.pow(hw[i,j],(5./3.)) * ti.math.pow(tSw,0.5)

	@ti.func
	def neighbour(i,j,k:int):
		ir,jr = -1,-1
		valid = True
		if(i == 0):
			if(j == 0 and k <= 1):
				valid = False
			elif(j == nx-1 and (k == 0 or k == 2)):
				valid = False
			elif(k==0):
				valid = False
		elif(j == 0 and k == 1):
			valid = False
		elif(j == nx-1 and k == 2):
			valid = False
		elif(i == ny-1):
			if(j == 0 and (k == 1 or k == 3)):
				valid = False
			elif(j == nx-1 and (k == 3 or k == 2)):
				valid = False
			elif(k==3):
				valid = False

		if(valid):
			if(k == 0):
				ir,jr = i-1, j
			if(k == 1):
				ir,jr = i, j-1
			if(k == 2):
				ir,jr = i, j+1
			if(k == 3):
				ir,jr = i+1, j
		return ir, jr



	@ti.kernel
	def compute_Qw():
		for i,j in Z:

			if(i == ny-1 or i ==0 or j==0 or j == nx-1 or Z[i,j] < 0.):
				continue

			Sws = ti.math.vec4(0.,0.,0.,0.)
			sumSw = 0.
			SSx = 0.
			SSy = 0.
			lockcheck = 0
			while(sumSw == 0.):
				lockcheck += 1
				for k in range(4):
					ir,jr = neighbour(i,j,k)
					if(ir == -1):
						continue

					tS = Sw(i,j,ir,jr)
					if(tS <= 0):
						continue

					if(k == 0 or k == 3):
						if(tS > SSy):
							SSy = tS
					else:
						if(tS > SSx):
							SSx = tS

					Sws[k] = tS
					sumSw += tS
				if(sumSw == 0.):
					hw[i,j] += 1e-4
				if(lockcheck > 10000):
					break

			gradSw = ti.math.sqrt(SSx*SSx + SSy*SSy)
			if(gradSw == 0):
				continue
			QwC[i,j] = dx/manning * ti.math.pow(hw[i,j], 5./3) *sumSw/ti.math.sqrt(gradSw)
			# print(QwC[i,j])
			for k in range(4):
				ir,jr = neighbour(i,j,k)
				if(ir == -1):
					continue
				ti.atomic_add(QwB[ir,jr], Sws[k]/sumSw * QwA[i,j])



	@ti.kernel
	def compute_hw():
		for i,j in Z:
			QwA[i,j] = QwB[i,j]
			if(i == ny-1 or i ==0 or j==0 or j == nx-1 or Z[i,j] < 0.):
				continue
			hw[i,j] = ti.math.max(0.,hw[i,j] + (QwA[i,j] - QwC[i,j]) * dt/(dx*dy) ) 



		# GUI setup

	init_field()

	thread = None

	def run_model(stop_event, mainwindow):
		i = 0
		while not stop_event.is_set():
			i+=1
			stepinit()
			compute_Qw()
			compute_hw()

			if(i % 1000 == 0):
				mainwindow.widgets['mainmap'].drapePlot.set_data(hw.to_numpy())
				mainwindow.widgets['mainmap'].canvas.draw()

	def valueChanged(newP, stop_event, mainwindow):
		# changing to
		# print("changing to ", newP * 1e-3/3600)
		global thread
		stop_event.set()
		thread.join()
		stop_event.clear()
		P[None] = np.float32(newP * 1e-3/3600)
		# Create and start the thread
		thread = threading.Thread(target=run_model, args=(stop_event, mainwindow,))
		thread.start()







	def init_model(main_window, guimap, stop_event, fname):
		global thread
		mainwindow.add_widget("mainmap", guimap.map_widget_from_fname_for_graphflood(fname), parentLayout = 'map')
		Z.from_numpy(mainwindow.widgets['mainmap'].env.grid.Z2D)

		# Create and start the thread
		thread = threading.Thread(target=run_model, args=(stop_event, mainwindow,))
		thread.start()




	# Ensure using the Qt6Agg backend with PySide6
	matplotlib.use('QtAgg')

	# Creating the application (required by Qt)
	app = QtWidgets.QApplication(sys.argv)

	# the mainwindow is the main Widget containing the others
	mainwindow = MainWindow(title = "HaGUId")

	main_style = """
		QWidget {
			background-color: #D3D3D3; /* Light Gray */
			border-style: solid;
			border-width: 2px;
			border-color: black;
		}

		QLabel {
			max-height: 20px;
			font-family: 'Courier New';
			font-size: 14pt;
			font-weight: normal;
			color: black;
			text-align: left;
			border: none;
		}

		QPushButton {
			border: 2px solid #8f8f91;
			border-radius: 6px;
			background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
											  stop: 0 #f6f7fa, stop: 1 #dadbde);
			min-width: 80px;
		}

		QPushButton:pressed {
			background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
											  stop: 0 #dadbde, stop: 1 #f6f7fa);
		}

		QPushButton:flat {
			border: none; /* no border for a flat push button */
		}

		QPushButton:default {
			border-color: navy; /* make the default button prominent */
		}

		QSlider {
			min-width: 50 px;
			border: none;
		}
		"""

	from scabbard.gui_stylesheets import dark_orange
	main_style = dark_orange

	##############################################################################################################################
	##############################################################################################################################
	# Starting by placing placeholders and their respective styles
	mainwindow.add_layout('menu', QtWidgets.QHBoxLayout(), 0,0, style = main_style)

	mainwindow.add_layout('toolbox', QtWidgets.QHBoxLayout(), 0,1, style = main_style)

	mainwindow.add_layout('map', QtWidgets.QHBoxLayout(), 1,0, style = main_style)

	mainwindow.add_layout('toolopt', QtWidgets.QHBoxLayout(), 1,1, style = main_style)

	# ttoolbox = QtWidgets.QLabel("toolopt")
	# mainwindow.layouts['toolopt'].addWidget(ttoolbox)
	ttoolopt = QtWidgets.QLabel("Current options: set precipitation rates")
	mainwindow.layouts['toolbox'].addWidget(ttoolopt)

	# Adjusting the stretch factors to control the fractions
	# These factors do not correspond directly of percentage of area but affect the distribution of space
	mainwindow.grid_layout.setColumnStretch(0, 5) # 30% width approximately
	mainwindow.grid_layout.setColumnStretch(1, 3) # 20% width approximately
	mainwindow.grid_layout.setRowStretch(0, 1) # 30% height approximately
	mainwindow.grid_layout.setRowStretch(1, 5) # 20% height approximately
	##############################################################################################################################
	##############################################################################################################################

	def add_scroll_bars(mainwindow):
		mainwindow.add_widget( 'P_slider_lab',QtWidgets.QLabel("Precipitation rates (1-300 mm/yrs)"),parentLayout = 'toolopt')
		mainwindow.add_widget( 'P_slider',FloatSlider(QtCore.Qt.Vertical,vmin = 1,vmax = 300, step = 1, multiplier = 50, label = "Precipitation rates"),parentLayout = 'toolopt')
		mainwindow.widgets['P_slider'].valMod.connect(lambda x: valueChanged(x,stop_event, mainwindow))


	mainwindow.add_widget("loader", FileLoader(),  parentLayout = 'menu')
	mainwindow.widgets['loader'].theChosenFile.connect(lambda x: (init_model(mainwindow, guimap, stop_event,x), add_scroll_bars(mainwindow) ,None)[-1])



	mainwindow.show()
	sys.exit(app.exec_())





# stop_event.set()

# # Wait for the thread to finish
# thread.join()






























# end of file